#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **callable parameter iterator utilities** (i.e., low-level
callables introspectively iterating over parameters accepted by arbitrary
callables).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilCallableException
from beartype.typing import (
    Iterable,
    Optional,
    Tuple,
)
from beartype._data.typing.datatyping import TypeException
from beartype._data.kind.datakindiota import Iota
from beartype._data.kind.datakindmap import FROZENDICT_EMPTY
from beartype._util.func.arg.utilfuncarglen import get_func_args_lens
from collections.abc import Callable
from enum import (
    Enum,
    auto as next_enum_member_value,
    unique as die_unless_enum_member_values_unique,
)
from itertools import count
from types import CodeType

# ....................{ ENUMERATIONS                       }....................
@die_unless_enum_member_values_unique
class ArgKind(Enum):
    '''
    Enumeration of all kinds of **callable parameters** (i.e., arguments passed
    to pure-Python callables).

    This enumeration intentionally declares members of the same name as those
    declared by the standard :class:`inspect.Parameter` class. Whereas the
    former are unconditionally declared below and thus portable across Python
    versions, the latter are only conditionally declared depending on Python
    version and thus non-portable across Python versions. Notably, the
    :attr:`inspect.Parameter.POSITIONAL_ONLY` attribute is only defined under
    Python >= 3.8.

    Attributes
    ----------
    POSITIONAL_ONLY : EnumMemberType
        Kind of all **positional-only parameters** (i.e., parameters required
        to be passed positionally, syntactically followed in the signatures of
        their callables by the :pep:`570`-compliant ``/,`` pseudo-parameter).
    POSITIONAL_OR_KEYWORD : EnumMemberType
        Kind of all **flexible parameters** (i.e., parameters permitted to be
        passed either positionally or by keyword).
    VARIADIC_POSITIONAL : EnumMemberType
        Kind of all **variadic positional parameters** (i.e., tuple of zero or
        more positional parameters *not* explicitly named by preceding
        positional-only or flexible parameters, syntactically preceded by the
        ``*`` prefix and typically named ``*args``).
    KEYWORD_ONLY  : EnumMemberType
        Kind of all **keyword-only parameters** (i.e., parameters required to
        be passed by keyword, syntactically preceded in the signatures of
        their callables by the :pep:`3102`-compliant ``*,`` pseudo-parameter).
    VARIADIC_KEYWORD : EnumMemberType
        Kind of all **variadic keyword parameters** (i.e., tuple of zero or
        more keyword parameters *not* explicitly named by preceding
        keyword-only or flexible parameters, syntactically preceded by the
        ``**`` prefix and typically named ``**kwargs``).
    '''

    POSITIONAL_ONLY = next_enum_member_value()
    POSITIONAL_OR_KEYWORD = next_enum_member_value()
    VARIADIC_POSITIONAL = next_enum_member_value()
    KEYWORD_ONLY = next_enum_member_value()
    VARIADIC_KEYWORD = next_enum_member_value()

# ....................{ SINGLETONS                         }....................
ArgMandatory = Iota()
'''
Arbitrary sentinel singleton assigned by the :func:`.iter_func_args` generator
to the tuple item with index :data:`.ARG_META_INDEX_DEFAULT` of all
:class:`.ArgMeta` instances, describing **mandatory parameters** (i.e.,
parameters that *must* be explicitly passed to their callables).
'''

# ....................{ HINTS                              }....................
ArgMeta = Tuple[ArgKind, str, object]
'''
PEP-compliant type hint matching each **callable parameter metadata**
iteratively yielded by the :func:`.iter_func_args` generator for each parameter
accepted by the pure-Python callable passed to that generator, defined as the
4-tuple ``(arg_kind, arg_name, arg_default_or_mandatory, arg_hint_or_unhinted)``
where:

* ``arg_kind`` is this parameter's **kind** (i.e., :class:`.ArgKind` enumeration
  member conveying the syntactic class of this parameter, constraining how the
  callable declaring this parameter requires this parameter to be passed).
* ``arg_name`` is this parameter's **name** (i.e., syntactically valid Python
  identifier uniquely identifying this parameter in its parameter list).
* ``arg_default_or_mandatory`` is either:

  * If this parameter is mandatory, the magic constant :data:`.ArgMandatory`.
  * Else, this parameter is optional and thus defaults to a default value when
    unpassed. In this case, this is that default value.

Caveats
-------
**This metadata intentionally omits the type hint annotating this parameter.**
Why? Because deciding *which* type hint annotates this parameter (if any) is
non-trivial in common edge cases. Notably, if the pure-Python callable passed to
that generator differs from the callable to be unwrapped with respect to type
hints, then that generator *cannot* reliably decide which type hint annotates
this parameter without refactoring both that generator and all other callables
transitively calling that generator to also accept and forward on a new
``type_hint_wrapper`` parameter providing the callable to be unwrapped with
respect to type hints. Although feasible, doing so would only invite even
further API confusion with little to no tangible benefit. Edge cases in which
this arises include:

* When ``type_hint_wrapper`` is a **pseudo-callable** (i.e., otherwise
  uncallable object whose type renders that object callable by defining the
  ``__call__()`` dunder method) *and* ``func`` is that ``__call__()`` dunder
  method. If that pseudo-callable wraps a lower-level callable, then that
  pseudo-callable (rather than ``__call__()`` dunder method) defines the
  ``__wrapped__`` instance variable providing that callable.

The caller is thus responsible for mapping parameters to type hints. Thankfully,
the existing
:attr:`beartype._check.meta.checkdecor.BeartypeDecorMeta.func_annotations`
dictionary makes this trivial for most use cases.
'''

# ....................{ CONSTANTS ~ index                  }....................
# Iterator yielding the next integer incrementation starting at 0, to be safely
# deleted *AFTER* defining the following 0-based indices via this iterator.
__arg_meta_index_counter = count(start=0, step=1)


ARG_META_INDEX_KIND = next(__arg_meta_index_counter)
'''
0-based index into each 4-tuple iteratively yielded by the generator returned by
the :func:`.iter_func_args` generator function providing the currently iterated
parameter's **kind** (i.e., :class:`ArgKind` enumeration member conveying this
parameter's syntactic class, constraining how the callable declaring this
parameter requires this parameter to be passed).
'''


ARG_META_INDEX_NAME = next(__arg_meta_index_counter)
'''
0-based index into each 4-tuple iteratively yielded by the generator returned by
the :func:`.iter_func_args` generator function providing the currently iterated
parameter's **name** (i.e., syntactically valid Python identifier uniquely
identifying this parameter in its parameter list).
'''


ARG_META_INDEX_DEFAULT = next(__arg_meta_index_counter)
'''
0-based index into each 4-tuple iteratively yielded by the generator returned by
the :func:`.iter_func_args` generator function providing the currently iterated
parameter's **default value**, defined as either:

* If this parameter is mandatory, the magic constant :data:`.ArgMandatory`.
* Else, this parameter is optional and thus defaults to a default value when
  unpassed. In this case, this is that default value.
'''


# Delete the above counter for safety and sanity in equal measure.
del __arg_meta_index_counter

# ....................{ GENERATORS                         }....................
def iter_func_args(
    # Mandatory parameters.
    func: Callable,

    # Optional parameters.
    func_codeobj: Optional[CodeType] = None,
    is_omit_boundmethod_arg_first: bool = True,
    is_unwrap: bool = True,
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
# Note this generator is intentionally annotated as returning a high-level
# "Iterable[...]" rather than a low-level "Generator[..., ..., ...]", as the
# syntax governing the latter is overly verbose and largely unhelpful.
) -> Iterable[ArgMeta]:
    '''
    Generator yielding one **parameter metadata tuple** (i.e., tuple whose items
    describe a single parameter) for each parameter accepted by the passed
    pure-Python callable.

    For consistency with the official grammar for callable signatures
    standardized by :pep:`570`, this generator is guaranteed to yield parameter
    metadata in the same order as required by Python syntax and semantics. In
    order, this is:

    * **Mandatory positional-only parameters** (i.e., parameter metadata
      whose kind is :attr:`ArgKind.POSITIONAL_ONLY` and whose default value is
      :data:`ArgMandatory`).
    * **Optional positional-only parameters** (i.e., parameter metadata
      whose kind is :attr:`ArgKind.POSITIONAL_ONLY` and whose default value is
      *not* :data:`ArgMandatory`).
    * **Mandatory flexible parameters** (i.e., parameter metadata whose kind is
      :attr:`ArgKind.POSITIONAL_OR_KEYWORD` and whose default value is
      :data:`ArgMandatory`).
    * **Optional flexible parameters** (i.e., parameter metadata whose kind is
      :attr:`ArgKind.POSITIONAL_OR_KEYWORD` and whose default value is *not*
      :data:`ArgMandatory`).
    * **Variadic positional parameters** (i.e., parameter metadata whose kind
      is :attr:`ArgKind.VARIADIC_POSITIONAL` and whose default value is
      :data:`ArgMandatory`).
    * **Mandatory and optional keyword-only parameters** (i.e., parameter
      metadata whose kind is :attr:`ArgKind.KEYWORD_ONLY`). Unlike all other
      parameter kinds, keyword-only parameters are (by definition) unordered;
      ergo, Python explicitly permits mandatory and optional keyword-only
      parameters to be heterogeneously intermingled rather than clustered.
    * **Variadic keyword parameters** (i.e., parameter metadata whose kind
      is :attr:`ArgKind.VARIADIC_KEYWORD` and whose default value is
      :data:`ArgMandatory`).

    Caveats
    -------
    **This highly optimized generator function should always be called in lieu
    of the highly unoptimized** :func:`inspect.signature` **function,** which
    implements a similar introspection as this generator with significantly
    worse space and time consumption. Seriously. *Never* call that anywhere.

    Parameters
    ----------
    func : Callable
        Pure-Python callable to be inspected.
    func_codeobj: CodeType, optional
        Code object underlying that callable unwrapped. Defaults to
        :data:`None`, in which case this iterator internally defers to the
        comparatively slower :func:`.get_func_codeobj` function.
    is_omit_boundmethod_arg_first : bool, optional
        :data:`True` only if this generator implicitly omits the first mandatory
        flexible parameter accepted by that callable if that callable is a
        C-based bound method descriptor encapsulating either an instance method
        bound to an instance of a class *or* a class method bound to a class.
        Defaults to :data:`True`, instructing this generator to transparently
        yield the *actual* high-level parameters accepted by this bound method
        descriptor (rather than the low-level parameters accepted by the unbound
        method encapsulated by this bound method descriptor). While the default
        behaviour is typically desirable, valid use cases for the non-default
        behaviour do exist (e.g., crudely detecting the kind of that based on
        whether the unbound method encapsulated by this bound method descriptor
        accepts a first parameter named ``cls`` or ``self``).

        The default behaviour enables:

        * This generator to transparently support bound method descriptors,
          which then enables...
        * The private :func:`beartype._decor._nontype.decornontype.beartype_pseudofunc`
          decorator to type-check the bound ``__call__()`` method descriptor
          encapsulating the unbound ``__call__()`` dunder method defined on the
          class of pseudo-callable objects, which then enables...
        * The public :func:`beartype.beartype` decorator to type-check
          pseudo-callable objects.

        How? In this case, the aforementioned ``beartype_pseudofunc``
        decorator wraps this bound method descriptor with a dynamically
        generated wrapper function that does *not* accept a ``self`` or ``cls``
        parameter, since a bound method does *not* accept a ``self`` or ``cls``
        parameter. After all, that's why bound methods exist; they implicitly
        pass the instance or class to which they are bound as the value of the
        ``self`` or ``cls`` parameter to the unbound method they encapsulate.
        However, the code object of a bound method descriptor is only an alias
        to the code object of the corresponding unbound method. Since the latter
        accepts a ``self`` parameter, so too does the former.

        The default behaviour resolves this internal discrepancy (contradiction)
        that arises between:

        * The code object of a bound method descriptor, which declares that
          callable object to accept a ``self`` parameter.
        * The real-world calling semantics of a bound method descriptor, which
          by definition accepts *no* ``self`` parameter.
    is_unwrap : bool, optional
        :data:`True` only if this generator implicitly calls the
        :func:`unwrap_func_all_isomorphic` function to unwrap this possibly
        higher-level wrapper into its possibly lowest-level wrappee *before*
        returning the code object of that wrappee. Note that doing so incurs
        worst-case time complexity ``O(n)`` for ``n`` the number of lower-level
        wrappees wrapped by this wrapper. Defaults to :data:`True` for
        robustness. Why? Because this generator *must* always introspect
        lowest-level wrappees rather than higher-level wrappers. The latter
        typically do *not* wrap the default values of the former, since this is
        the default behaviour of the :func:`functools.update_wrapper` function
        underlying the :func:`functools.wrap` decorator underlying all sane
        decorators. If this boolean is set to :data:`False` while that callable
        is actually a wrapper, this generator will erroneously misidentify
        optional as mandatory parameters and fail to yield their default values.
        Only set this boolean to :data:`False` if you pretend to know what
        you're doing.
    exception_cls : type, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :class:`._BeartypeUtilCallableException`.
    exception_prefix : str, optional
        Human-readable label prefixing the message of any exception raised in
        the event of a fatal error. Defaults to the empty string.

    Yields
    ------
    ArgMeta
        Parameter metadata tuple describing the currently yielded parameter.

    Raises
    ------
    exception_cls
         If that callable is *not* pure-Python.
    '''
    assert isinstance(is_omit_boundmethod_arg_first, bool), (
        f'{repr(is_omit_boundmethod_arg_first)} not boolean.')
    assert isinstance(is_unwrap, bool), f'{repr(is_unwrap)} not boolean.'

    # ..................{ IMPORTS                            }..................
    # Avoid circular import dependencies.
    from beartype._util.func.utilfunccodeobj import get_func_codeobj
    from beartype._util.func.utilfunctest import is_func_boundmethod
    from beartype._util.func.utilfuncwrap import unwrap_func_all_isomorphic

    # ..................{ LOCALS ~ core                      }..................
    # If unwrapping that callable, do so *BEFORE* obtaining the code object of
    # that callable for safety (to avoid desynchronization between the two).
    if is_unwrap:
        func = unwrap_func_all_isomorphic(func)
    # Else, that callable is assumed to have already been unwrapped by the
    # caller. We should probably assert that, but doing so requires an
    # expensive call to hasattr(). What you gonna do?

    # Number of various kinds of parameters accepted by that callable.
    (
        # Number of both optional and mandatory non-keyword-only parameters
        # (i.e., positional-only *AND* flexible (i.e., positional or keyword)
        # parameters) accepted by that callable.
        args_len_posonly_or_flex,
        # Number of both optional and mandatory keyword-only parameters accepted
        # by that callable.
        args_len_kwonly,
        # 1 only if that callable accepts variadic positional or keyword
        # parameters and 0 otherwise.
        is_arg_var_pos,
        is_arg_var_kw,
    ) = get_func_args_lens(
        func=func,
        is_unwrap=False,  # <-- "func" was already unwrapped above. I sigh.
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # If that callable accepts *NO* parameters, silently reduce to the empty
    # generator (i.e., noop) for both space and time efficiency. Just. Do. It.
    #
    # Note that this is a critical optimization for the common use case that
    # @beartype is unconditionally applied with "beartype.claw"-based import
    # hook automation to *ALL* callables physically declared by a package, many
    # of which will inevitably be argumentless.
    if (
        args_len_posonly_or_flex +
        args_len_kwonly +
        is_arg_var_pos +
        is_arg_var_kw
    ) == 0:
        yield from ()
        return
    # Else, that callable accepts one or more parameters.

    # If passed *NO* code object, query that callable for its code object.
    if func_codeobj is None:
        func_codeobj = get_func_codeobj(
            func=func,
            is_unwrap=False,  # <-- "func" was already unwrapped above. I sigh.
            exception_cls=exception_cls,
            exception_prefix=exception_prefix,
        )
    # In any case, that code object is now defined.

    # Tuple of the names of all variables localized to that callable.
    #
    # Note that this tuple contains the names of both:
    # * All parameters accepted by that callable.
    # * All local variables internally declared in that callable's body.
    #
    # Ergo, this tuple *CANNOT* be searched in full. Only the subset of this
    # tuple containing argument names is relevant and may be safely searched.
    #
    # Lastly, note the "func_codeobj.co_names" attribute is incorrectly
    # documented in the "inspect" module as the "tuple of names of local
    # variables." That's a lie. Instead, that attribute is a mostly useless
    # tuple of the names of both globals and object attributes accessed in the
    # body of that callable. *shrug*
    args_name = func_codeobj.co_varnames

    # ..................{ LOCALS ~ defaults                  }..................
    # Tuple of the default values assigned to all optional non-keyword-only
    # parameters (i.e., all optional positional-only *AND* optional flexible
    # (i.e., positional or keyword) parameters) accepted by that callable if any
    # *OR* the empty tuple otherwise.
    args_defaults_posonly_or_flex = func.__defaults__ or ()  # type: ignore[attr-defined]
    # print(f'args_defaults_posonly_or_flex: {args_defaults_posonly_or_flex}')

    # Dictionary mapping from the name of each optional keyword-only parameter
    # accepted by that callable to the default value assigned to that parameter
    # if any *OR* the empty dictionary otherwise.
    #
    # For both space and time efficiency, the empty dictionary is intentionally
    # *NOT* accessed here as "{}". Whereas each instantiation of the empty tuple
    # efficiently reduces to the same empty tuple, each instantiation of the
    # empty dictionary inefficiently creates a new empty dictionary: e.g.,
    #     >>> () is ()
    #     True
    #     >>> {} is {}
    #     False
    args_defaults_kwonly = func.__kwdefaults__ or FROZENDICT_EMPTY  # type: ignore[attr-defined]

    # ..................{ LOCALS ~ len                       }..................
    # Number of both optional and mandatory positional-only parameters accepted
    # by that callable, standardized under Python >= 3.8 by PEP 570.
    args_len_posonly = func_codeobj.co_posonlyargcount  # type: ignore[attr-defined]
    assert args_len_posonly_or_flex >= args_len_posonly, (
        f'Positional-only and flexible argument count {args_len_posonly_or_flex} < '
        f'positional-only argument count {args_len_posonly}.')

    # Number of both optional and mandatory flexible parameters accepted by
    # that callable.
    args_len_flex = args_len_posonly_or_flex - args_len_posonly

    # Number of optional non-keyword-only parameters accepted by that callable.
    args_len_posonly_or_flex_optional = len(args_defaults_posonly_or_flex)

    # Number of optional flexible parameters accepted by that callable, defined
    # as the number of optional non-keyword-only parameters capped to the total
    # number of flexible parameters. Why? Because optional flexible parameters
    # preferentially consume non-keyword-only default values first; optional
    # positional-only parameters consume all remaining non-keyword-only default
    # values. Why? Because:
    # * Default values are *ALWAYS* assigned to positional parameters from
    #   right-to-left.
    # * Flexible parameters reside to the right of positional-only parameters.
    #
    # Specifically, this number is defined as...
    args_len_flex_optional = min(
        # If the number of optional non-keyword-only parameters exceeds the
        # total number of flexible parameters, the total number of flexible
        # parameters. For obvious reasons, the number of optional flexible
        # parameters *CANNOT* exceed the total number of flexible parameters;
        args_len_flex,
        # Else, the total number of flexible parameters is strictly greater
        # than the number of optional non-keyword-only parameters, implying
        # optional flexible parameters consume all non-keyword-only default
        # values. In this case, the number of optional flexible parameters is
        # the number of optional non-keyword-only parameters.
        args_len_posonly_or_flex_optional,
    )

    # Number of optional positional-only parameters accepted by that callable,
    # defined as all remaining optional non-keyword-only parameters *NOT*
    # already consumed by positional parameters. Note that this number is
    # guaranteed to be non-negative. Why? Because, it is the case that either:
    # * "args_len_posonly_or_flex_optional >= args_len_flex", in which case
    #   "args_len_flex_optional == args_len_flex", in which case
    #   "args_len_posonly_or_flex_optional >= args_len_flex_optional".
    # * "args_len_posonly_or_flex_optional < args_len_flex", in which case
    #   "args_len_flex_optional == args_len_posonly_or_flex_optional", in which
    #   case "args_len_posonly_or_flex_optional == args_len_flex_optional".
    #
    # Just roll with it, folks. It's best not to question the unfathomable.
    args_len_posonly_optional = (
        args_len_posonly_or_flex_optional - args_len_flex_optional)

    # Number of mandatory positional-only parameters accepted by that callable.
    args_len_posonly_mandatory = args_len_posonly - args_len_posonly_optional

    # Number of mandatory flexible parameters accepted by that callable.
    args_len_flex_mandatory = args_len_flex - args_len_flex_optional

    # ..................{ ITERATE ~ positional-only          }..................
    # 0-based index of the first parameter of the currently iterated kind
    # accepted by that callable in the "args_name" tuple.
    args_index_kind_first = 0

    # If that callable accepts at least one mandatory positional-only
    # parameter...
    if args_len_posonly_mandatory:
        # For each mandatory positional-only parameter accepted by that
        # callable, yield a tuple describing this parameter.
        for arg_name in args_name[
            args_index_kind_first:args_len_posonly_mandatory]:
            yield (ArgKind.POSITIONAL_ONLY, arg_name, ArgMandatory,)

        # 0-based index of the first parameter of the next iterated kind.
        args_index_kind_first = args_len_posonly_mandatory
    # Else, that callable accepts *NO* mandatory positional-only parameters.

    # If that callable accepts at least one optional positional-only
    # parameter...
    if args_len_posonly_optional:
        # 0-based index of the parameter following the last optional
        # positional-only parameter in the "args_name" tuple.
        args_index_kind_last_after = (
            args_index_kind_first + args_len_posonly_optional)

        # For the 0-based index of each optional positional-only parameter
        # accepted by that callable and that parameter, yield a tuple
        # describing this parameter.
        for arg_index, arg_name in enumerate(args_name[
            args_index_kind_first:args_index_kind_last_after]):
            # assert arg_posonly_optional_index < args_len_posonly_optional, (
            #     f'Optional positional-only parameter index {arg_posonly_optional_index} >= '
            #     f'optional positional-only parameter count {args_len_posonly_optional}.')
            yield (
                ArgKind.POSITIONAL_ONLY,
                arg_name,
                args_defaults_posonly_or_flex[arg_index],
            )

        # 0-based index of the first parameter of the next iterated kind.
        args_index_kind_first = args_index_kind_last_after
    # Else, that callable accepts *NO* optional positional-only parameters.

    # ..................{ ITERATE ~ flexible                 }..................
    # If that callable accepts at least one mandatory flexible parameter...
    if args_len_flex_mandatory:
        # 0-based index of the parameter following the last mandatory flexible
        # parameter in the "args_name" tuple.
        args_index_kind_last_after = (
            args_index_kind_first + args_len_flex_mandatory)

        # If...
        if (
            # Omitting the first mandatory flexible parameter accepted by that
            # callable if that callable is a C-based bound method descriptor
            # encapsulating either an instance method bound to an instance of a
            # class or a class method bound to a class *AND*...
            is_omit_boundmethod_arg_first and
            # That callable is such a C-based bound method descriptor...
            is_func_boundmethod(func)
        ):
            # print(f'Ignoring bound method {repr(func)} first argument...')
            # Increment the 0-based index of the first mandatory flexible
            # parameter accepted by this method in the "args_name" tuple to
            # account for the first mandatory flexible "self" parameter
            # implicitly passed by this bound method descriptor to this method,
            # effectively ignoring this "self" parameter.
            #
            # Note that:
            # * We intentionally increment this index *AFTER* computing the
            #   derivative "args_index_kind_last_after" index above with the
            #   original value of this index.
            # * Handling this common edge case enables:
            #   * This generator to transparently support bound method
            #     descriptors, which then enables...
            #   * The private @beartype._decor._nontype.decornontype.beartype_pseudofunc
            #     decorator to type-check the bound __call__() method descriptor
            #     encapsulating the unbound __call__() dunder method defined on
            #     the class of pseudo-callable objects, which then enables...
            #   * The public @beartype.beartype decorator to type-check
            #     pseudo-callable objects.
            #
            #   How? In this case, the aforementioned @beartype_pseudofunc
            #   decorator wraps this bound method descriptor with a dynamically
            #   generated wrapper function that does *NOT* accept a "self"
            #   parameter, since a bound method does *NOT* accept a "self"
            #   parameter. However, the code object of a bound method descriptor
            #   is simply an alias of the code object of the corresponding
            #   unbound method. Since the latter accepts a "self" parameter, so
            #   too does the former. Thus, an internal discrepancy (arguably,
            #   contradiction) arises between:
            #
            #   * The code object of a bound method descriptor, which declares
            #     that callable object to accept a "self" parameter.
            #   * The real-world calling semantics of a bound method descriptor,
            #     which by definition accepts *NO* "self" parameter.
            args_index_kind_first += 1
        # Else, that callable is *NOT* such a descriptor (and is thus almost
        # certainly a vanilla pure-Python callable).

        # For each mandatory flexible parameter accepted by that callable, yield
        # a tuple describing this parameter.
        for arg_name in args_name[
            args_index_kind_first:args_index_kind_last_after]:
            yield (ArgKind.POSITIONAL_OR_KEYWORD, arg_name, ArgMandatory,)

        # 0-based index of the first parameter of the next iterated kind.
        args_index_kind_first = args_index_kind_last_after
    # Else, that callable accepts *NO* mandatory flexible parameters.

    # If that callable accepts at least one optional flexible parameter...
    if args_len_flex_optional:
        # 0-based index of the parameter following the last optional
        # flexible parameter in the "args_name" tuple.
        args_index_kind_last_after = (
            args_index_kind_first + args_len_flex_optional)

        # For the 0-based index of each optional flexible parameter accepted by
        # this callable and that parameter, yield a 3-tuple describing this
        # parameter.
        for arg_index, arg_name in enumerate(args_name[
            args_index_kind_first:args_index_kind_last_after]):
            # assert arg_flex_optional_index < args_len_flex_optional, (
            #     f'Optional flexible parameter index {arg_flex_optional_index} >= '
            #     f'optional flexible parameter count {args_len_flex_optional}.')
            yield (
                ArgKind.POSITIONAL_OR_KEYWORD,
                arg_name,
                args_defaults_posonly_or_flex[
                    args_len_posonly_optional + arg_index],
            )

        # 0-based index of the first parameter of the next iterated kind.
        args_index_kind_first = args_index_kind_last_after
    # Else, that callable accepts *NO* optional flexible parameters.

    # ..................{ ITERATE ~ variadic positional      }..................
    # 0-based index of the parameter following the last keyword-only
    # parameter in the "args_name" tuple. This index is required by multiple
    # branches below (rather than merely one branch) and thus unconditionally
    # computed for all these branches.
    args_index_kind_last_after = args_index_kind_first + args_len_kwonly

    # If that callable accepts a variadic positional parameter, yield a tuple
    # describing this parameter.
    #
    # Note that:
    # * This parameter is intentionally yielded *BEFORE* keyword-only
    #   parameters to conform with syntactic standards. A variadic positional
    #   parameter necessarily appears before any keyword-only parameters in the
    #   signature of that callable.
    # * The 0-based index of this parameter in the "args_name" tuple is exactly
    #   one *AFTER* the last keyword-only parameter in that tuple if any and
    #   one *BEFORE* the variadic keyword parameter in that tuple if any. This
    #   idiosyncrasy is entirely the fault of CPython, which grouped the
    #   two variadic positional and keyword parameters at the end of this list
    #   despite syntactic constraints on their lexical position.
    if is_arg_var_pos:
        yield (
            ArgKind.VARIADIC_POSITIONAL,
            args_name[args_index_kind_last_after],
            ArgMandatory,
        )
    # Else, that callable accepts *NO* variadic positional parameter.

    # ..................{ ITERATE ~ keyword-only             }..................
    # If that callable accepts at least one keyword-only parameter...
    if args_len_kwonly:
        # dict.get() method repeatedly called below and thus localized for
        # negligible efficiency. Look. Just do this. We needs godspeed.
        args_defaults_kwonly_get = args_defaults_kwonly.get

        # For each keyword-only parameter accepted by that callable, yield a
        # tuple describing this parameter.
        for arg_name in args_name[
            args_index_kind_first:args_index_kind_last_after]:
            yield (
                ArgKind.KEYWORD_ONLY,
                arg_name,
                # Either:
                # * If this is an optional keyword-only parameter, the default
                #   value of this parameter.
                # * If this is a mandatory keyword-only parameter, the
                #   placeholder "ArgMandatory" singleton.
                args_defaults_kwonly_get(arg_name, ArgMandatory),
            )
    # Else, that callable accepts *NO* keyword-only parameters.

    # ..................{ ITERATE ~ variadic keyword         }..................
    # If that callable accepts a variadic keyword parameter...
    if is_arg_var_kw:
        # 0-based index of the variadic keyword parameter accepted by that
        # callable in the "args_name" tuple, optimized by noting that Python
        # booleans are literally integers that can be computed with. Notably:
        # * If that callable accepts *NO* variadic positional parameter, then:
        #       is_arg_var_pos == 0
        #       args_index_var_kw == args_index_var_pos
        # * If that callable accepts a variadic positional parameter, then:
        #       is_arg_var_kw == 1
        #       args_index_var_pos == args_index_var_pos + 1
        args_index_kind_last_after += is_arg_var_pos

        # Yield a tuple describing this parameter.
        yield (
            ArgKind.VARIADIC_KEYWORD,
            args_name[args_index_kind_last_after],
            ArgMandatory,
        )
    # Else, that callable accepts *NO* variadic keyword parameter.



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`646`- and :pep:`692`-compliant **tuple type hint utilities**
(i.e., low-level callables generically applicable to :pep:`646`-compliant tuple
type hints *and* :pep:`692`-compliant unpacked typed dictionaries alike).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep646692Exception
from beartype._cave._cavefast import (
    HintGenericSubscriptedType,
    HintPep646TypeVarTupleType,
)
from beartype._data.typing.datatypingport import (
    Hint,
    TupleHints,
)
from beartype._data.hint.sign.datahintsigncls import HintSign
from beartype._data.hint.sign.datahintsigns import (
    HintSignPep646TypeVarTupleUnpacked,
    HintSignPep692TypedDictUnpacked,
)

# ....................{ RAISERS                            }....................
#FIXME: Unit test us up, please. *sigh*
def die_unless_hint_pep646_typevartuple_packed(hint: Hint) -> None:
    '''
    Raise an exception unless the passed hint is a :pep:`646`-compliant **packed
    type variable tuple** (i.e., :class:`TypeVarTuple` object).

    Parameters
    ----------
    hint : Hint
        Hint to be validated.

    Raises
    ------
    BeartypeDecorHintPep646692Exception
        If this hint is *not* a :pep:`646`-compliant packed type variable tuple.
    '''

    # If this hint is *NOT* a PEP 646-compliant type variable tuple, raise an
    # exception.
    if not isinstance(hint, HintPep646TypeVarTupleType):
        raise BeartypeDecorHintPep646692Exception(
            f'Type hint {repr(hint)} not PEP 646 type variable tuple '
            f'(i.e., "typing.TypeVarTuple" object).'
        )
    # Else, this hint is a PEP 646-compliant type variable tuple.

# ....................{ TESTERS                            }....................
#FIXME: *INSUFFICIENT.* Technically, it's true that *MOST* real-world unpacked
#tuples are low-level C objects created with the "*" unary prefix and thus
#correctly matched by this tester. That said, *SOME* real-world unpacked
#tuples are high-level pure-Python objects created by the "typing.Unpack[...]"
#hint factory and thus incorrectly *NOT* matched by this tester.
#
#Notably, PEP 696 explicitly insists that the unpacked tuple hint
#"Unpack[tuple[str, int]]" is semantically equivalent to the unpacked tuple hint
#"*tuple[str, int]":
#    DefaultTs = TypeVarTuple("DefaultTs", default=Unpack[tuple[str, int]])
#
#That said, it could be tricky to attempt to match both with this low-level
#tester function. Since this tester is *ONLY* called by
#get_hint_pep_sign_or_none(), it'd be better to go the disambiguation route.
#Notably:
#* Generalize the disambiguate_hint_pep646692_unpacked_sign() function defined
#  below to additionally support unpacked tuple hints. *shrug*
#FIXME: Unit test us up, please.
def is_hint_pep646_tuple_unpacked_prefix(hint: Hint) -> bool:
    '''
    :data:`True` only if the passed hint is a :pep:`646`-compliant **prefix-based
    unpacked child tuple hint** (i.e., of the form "*tuple[{hint_child_child_1},
    ..., {hint_child_child_M}]" subscripting a parent tuple hint of the form
    "tuple[{hint_child_1}, ..., *tuple[{hint_child_child_1}, ...,
    {hint_child_child_M}], ..., {hint_child_N}]").

    If this tester returns :data:`True`, this unpacked child tuple hint is
    guaranteed to define the ``__args__`` dunder attribute to be either:

    * A 2-tuple ``({hint_child}, ...)``, in which case this child tuple hint
      unpacks to a variable-length tuple hint over ``{hint_child}`` types.
    * An n-tuple ``({hint_child_1}, ..., {hint_child_N})`` where ``...`` in this
      case is merely a placeholder connoting one or more child hints, in which
      case this child tuple hint unpacks to a fixed-length tuple hint over these
      exact ``{hint_child_I}`` types.

    This getter is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as the implementation trivially reduces to
    an efficient one-liner.

    Motivation
    ----------
    Interestingly, even detecting accursed unpacked child tuple hints at runtime
    is highly non-trivial. They do *not* have a sane unambiguous type,
    significantly complicating detection. For some utterly inane reason, their
    type is simply the ambiguous type :class:`types.GenericAlias` (i.e.,
    :class:`.HintGenericSubscriptedType`). That... wasn't what we were expecting
    *at all*. For example, under Python 3.13:

    .. code-block:: python

       # Note that Python *REQUIRES* unpacked tuple type hints to be embedded in
       # some larger syntactic construct. So, just throw it into a list. This is
       # insane, because we're only going to rip it right back out of that list.
       # Blame the CPython interpreter. *shrug*
       >>> yam = [*tuple[int, str]]

       # Arbitrary unpacked tuple type hint.
       >>> yim = yam[0]
       >>> repr(yim)
       *tuple[int, str]  # <-- gud
       >>> type(yim)
       <class 'types.GenericAlias'>  # <-- *TOTALLY NOT GUD. WTF, PYTHON!?*

       # Now look at this special madness. The type of this object isn't even in
       # its method-resolution order (MRO)!?!? I've actually never seen that
       # before. The type of any object is *ALWAYS* the first item in its
       # method-resolution order (MRO), isn't it? I... guess not. *facepalm*
       >>> yim.__mro__
       (<class 'tuple'>, <class 'object'>)

       # So, "*tuple[int, str]" is literally both a tuple *AND* a "GenericAlias"
       # at the same time. That makes no sense, but here we are. What are the
       # contents of this unholy abomination?
       >>> dir(yim)
       ['__add__', '__args__', '__bases__', '__class__', '__class_getitem__',
       '__contains__', '__copy__', '__deepcopy__', '__delattr__', '__dir__',
       '__doc__', '__eq__', '__format__', '__ge__', '__getattribute__',
       '__getitem__', '__getnewargs__', '__getstate__', '__gt__', '__hash__',
       '__init__', '__init_subclass__', '__iter__', '__le__', '__len__', '__lt__',
       '__mro_entries__', '__mul__', '__ne__', '__new__', '__origin__',
       '__parameters__', '__reduce__', '__reduce_ex__', '__repr__', '__rmul__',
       '__setattr__', '__sizeof__', '__str__', '__subclasshook__',
       '__typing_unpacked_tuple_args__', '__unpacked__', 'count', 'index']

       # Lotsa weird stuff there, honestly. Let's poke around a bit.
       >>> yim.__args__
       (<class 'int'>, <class 'str'>)  # <-- gud
       >>> yim.__typing_unpacked_tuple_args__
       (<class 'int'>, <class 'str'>)  # <-- gud, albeit weird
       >>> yim.__unpacked__
       True  # <-- *WTF!?!? what the heck is this nonsense?*
       >>> yim.__origin__
       tuple  # <-- so you lie about everything, huh?

    Basically, the above means that the only means of reliably detecting an
    unpacked tuple hint at runtime is by heuristically introspecting for both
    the existence *and* values of various dunder attributes exhibited above.
    Introspecting merely the existence of these attributes is insufficient; only
    the combination of both existence and values suffices to effectively
    guarantee disambiguity. Likewise, introspecting merely one or even two of
    these attributes is insufficient; only the combination of three or more of
    these attributes suffices to effectively guarantee disambiguity.

    Note there *are* no means of actually guaranteeing disambiguity. Malicious
    third-party objects could attempt to masquerade as unpacked child tuple
    hints by defining similar dunder attributes. We can only reduce the
    likelihood of false positives by increasing the number of dunder attributes
    introspected by this tester. Don't blame us. We didn't start the fire.

    Parameters
    ----------
    hint : Hint
        Type hint to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this hint is an unpacked child tuple hint.
    '''
    # print(f'dir({hint}): {dir(hint)}')
    # print(f'{hint}.__class__: {hint.__class__}')
    # print(f'{hint}.__args__: {getattr(hint, '__args__', False)}')
    # print(f'{hint}.__typing_unpacked_tuple_args__: {getattr(hint, '__typing_unpacked_tuple_args__', True)}')

    # Return true only if...
    return (
        # This hint's type is that of all unpacked child tuple hints (as well as
        # many other unrelated first- and third-party hints, sadly) *AND*...
        hint.__class__ is HintGenericSubscriptedType and
        (
            # The tuple of all child hints subscripting this parent hint is also
            # the tuple of all child hints subscripting this unpacked child
            # tuple hint. Since only unpacked child tuple hints *SHOULD* define
            # the PEP 646-compliant and frankly outrageously verbose
            # "__typing_unpacked_tuple_args__" dunder attribute, this
            # equivalence *SHOULD* suffice to disambiguate this hint as an
            # unpacked child tuple hint.
            getattr(hint, '__args__', False) ==
            getattr(hint, '__typing_unpacked_tuple_args__', True)
        )
        #FIXME: Probably unnecessary for the moment. Let's avoid violating
        #fragile privacy encapsulation any more than we must, please. *sigh*
        # # This hint's method-resolution order (MRO) is that of all unpacked
        # # child tuple hints *AND*...
        # getattr(hint, '__mro__', None) == _PEP646_HINT_TUPLE_UNPACKED_MRO and  # pyright: ignore
        # # This hint defines a dunder attribute uniquely defined *ONLY* by
        # # unpacked child tuple hints with a value guaranteed to be set by all
        # # unpacked child tuple hints.
        # getattr(hint, '__unpacked__', None) is True
    )

# ....................{ DISAMBIGUATORS                     }....................
#FIXME: Unit test us up, including:
#* "typing.Unpack[tuple[...]]" type hints. See discussion above. *sigh*
#* Unsubscripted "typing.Unpack" type hints, which should be unconditionally
#  *PROHIBITED.* They signify nothing. "typing.Unpack" should *ALWAYS* be
#  subscripted by at least something.
def disambiguate_hint_pep646692_unpacked_sign(hint: Hint) -> HintSign:
    '''
    Disambiguate the passed **unpacked type hint** (i.e., :pep:`646`- or
    :pep:`692`-compliant ``typing.Unpack[...]`` hint) ambiguously identified by
    the :data:`.HintSignUnpack` sign into whichever of the unambiguous
    :data:`.HintSignPep646TypeVarTupleUnpacked` or
    :data:`.HintSignPep692TypedDictUnpacked` signs uniquely identify this kind
    of unpacked type hint.

    This low-level getter assists the higher-level
    :func:`beartype._util.hint.pep.utilpepget.get_hint_pep_sign_or_none` getter
    to disambiguate the originally ambiguous :data:`.HintSignUnpack` sign.

    This low-level getter is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as only function calling this getter is the
    aforementioned higher-level getter, which itself is memoized.

    Parameters
    ----------
    hint : Hint
        Unpacked type hint to be disambiguated.

    Returns
    -------
    HintSign
        Sign uniquely and unambiguously identifying this hint. Specifically, if
        this hint is a:

        * :pep:`646`-compliant **unpacked type variable tuple** (i.e., child
          hint of the form "*{typevartuple}" where "{typevartuple}" is an
          instance of the :class:`typing.TypeVarTuple` type), this getter
          returns :data:`.HintSignPep646TypeVarTupleUnpacked`.
        * :pep:`692`-compliant **unpacked typed dictionary** (i.e., hint of the
          form "*{typeddict}" where "{typeddict}" is an instance of the
          :class:`typing.TypedDict` type factory), this getter
          returns :data:`.HintSignPep692TypedDictUnpacked`.

    Raises
    ------
    BeartypeDecorHintPep646692Exception
        If this hint is neither :pep:`646`- nor :pep:`692`-compliant (i.e., is a
        ``typing.Unpack[...]`` hint subscripted by a child hint that is neither
        an unpacked type variable tuple nor unpacked typed dictionary).
    '''
    # Note that this lower-level getter is directly called by the higher-level
    # get_hint_pep_sign_or_none() getter. Ergo, the former *CANNOT* recursively
    # pass the passed hint to the latter (e.g., as a means of validating that
    # the passed hint is indeed a "typing.Unpack[...]" hint). This getter *MUST*
    # assume the caller to pass a "typing.Unpack[...]" hint. It is what it is.
    # print(f'Disambiguating unpack hint {repr(hint)}...')

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype._util.hint.pep.proposal.pep589 import is_hint_pep589
    from beartype._util.hint.pep.utilpepget import get_hint_pep_args
    from beartype._util.hint.utilhinttest import die_as_hint_unsupported

    # ....................{ LOCALS                         }....................
    # Child hints subscripting this parent tuple hint.
    hint_childs = get_hint_pep_args(hint)
    # print(f'hint_childs: {hint_childs}')

    # If this parent unpack hint is subscripted by exactly one child hint...
    #
    # Note that the "typing.Unpack" type hint factory has already pre-validated
    # this factory to accept at most one child type hint. Nonetheless, one can
    # never be too careful where the "typing" module is concerned:
    #     >>> from typing import Unpack
    #     >>> Unpack['shaking', 'my', 'head']
    #     TypeError: typing.Unpack accepts only single type. Got ('shaking',
    #     'my', 'head').
    if len(hint_childs) == 1:
        # Child hint subscripting this parent unpack hint.
        hint_child = hint_childs[0]

        # If this child hint is a PEP 646-compliant type variable tuple, return
        # the sign disambiguating this parent unpack hint as a PEP 646-compliant
        # unpacked type variable tuple.
        if isinstance(hint_child, HintPep646TypeVarTupleType):
            return HintSignPep646TypeVarTupleUnpacked
        # Else, this child hint is *NOT* a PEP 646-compliant unpacked type
        # variable tuple.
        #
        # If this child hint is a PEP 589-compliant typed dictionary, return
        # the sign disambiguating this parent unpack hint as a PEP 692-compliant
        # unpacked typed dictionary.
        elif is_hint_pep589(hint_child):
            return HintSignPep692TypedDictUnpacked
        # Else, this child hint is *NOT* a PEP 589-compliant typed dictionary.
    # Else, this parent unpack hint is subscripted by either no child hints *OR*
    # two or more child hints.

    # ....................{ EXCEPTION                      }....................
    # Raise an exception. The child hint subscripting this parent unpack hint is
    # unrecognized and thus either PEP-noncompliant *OR* PEP-compliant but
    # unsupported by beartype. In either case, the caller deserves to know.
    die_as_hint_unsupported(
        hint=hint, exception_cls=BeartypeDecorHintPep646692Exception)

# ....................{ FACTORIES ~ tuple                  }....................
def make_hint_pep646_tuple_unpacked_prefix(hints_child: TupleHints) -> Hint:
    '''
    Dynamically create and return a new :pep:`646`-compliant **prefix-based
    unpacked child tuple hint** (i.e., of the form "*tuple[{hint_child_child_1},
    ..., {hint_child_child_M}]" subscripting a parent tuple hint of the form
    "tuple[{hint_child_1}, ..., *tuple[{hint_child_child_1}, ...,
    {hint_child_child_M}], ..., {hint_child_N}]") subscripted by all child hints
    in the passed tuple.

    This factory exists to streamline access to unpacked child tuple hints,
    whose definition is otherwise non-trivial. Python requires unpacked child
    tuple hints to be syntactically embedded inside larger containers -- even if
    those hints are semantically invalid inside those containers: e.g.,

    .. code-block:: pycon

       >>> [*tuple[int, str]]
       [*tuple[int, str]]  # <-- makes no sense, but ok.
       >>> *tuple[int, str]
       SyntaxError: can't use starred expression here  # <-- this is awful.

    This factory circumvents these non-trivial usability concerns.

    Parameters
    ----------
    hints_child : TupleHints
        Tuple of all child hints to be unpacked.

    Returns
    -------
    Hint
        Unpacked child tuple hint subscripted by these child hints.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.proposal.pep484585646 import (
        make_hint_pep484585_tuple_fixed)

    # PEP 585-compliant tuple hint subscripted by these child hints.
    pep585_tuple = make_hint_pep484585_tuple_fixed(hints_child)

    #FIXME: Uncomment after dropping Python <= 3.10 support, which raises a
    #"SyntaxError" if we even try doing this. *SADNESS*
    # return *hint

    # Tuple subscripted by this PEP 646-compliant type tuple hint unpacked into
    # a PEP 646-compliant unpacked tuple child hint. This is insane, because
    # we're only going to rip this tuple child hint right back out of this
    # tuple. Blame the Python <= 3.10 interpreter.
    #
    # Note that a tuple was intentionally chosen for both space and time
    # efficiency. Although *ANY* container would satisfy Python <= 3.10, a tuple
    # has the advantage of being Python's most optimized container type.
    tuple_unpacked_parent = (*pep585_tuple,)  # type: ignore[valid-type]

    # PEP 646-compliant unpacked child tuple hint subscripting this parent.
    tuple_unpacked = tuple_unpacked_parent[0]

    # Return this unpacked child tuple hint.
    return tuple_unpacked

# ....................{ FACTORIES ~ typevartuple           }....................
#FIXME: Unit test us up, please. *sigh*
def make_hint_pep646_typevartuple_unpacked_prefix(
    hint: HintPep646TypeVarTupleType) -> Hint:
    '''
    Dynamically create and return a new :pep:`646`-compliant **prefix-based
    unpacked type variable tuple** (i.e., of the form ``*hint``) prefixing the
    passed type variable tuple by the unary unpacking ``*`` operator.

    Parameters
    ----------
    hint: HintPep646TypeVarTupleType
        Type variable tuple to be unpacked.

    Returns
    -------
    Hint
        Unpacked type variable tuple synthesized from this type variable tuple.

    See Also
    --------
    :func:`.make_hint_pep646_tuple_unpacked_prefix`
        Further discussion on prefix-based unpacking.
    '''

    # If this hint is *NOT* a type variable tuple, raise an exception.
    die_unless_hint_pep646_typevartuple_packed(hint)  # pyright: ignore
    # Else, this hint is a type variable tuple.

    #FIXME: Uncomment after dropping Python <= 3.10 support, which raises a
    #"SyntaxError" if we even try doing this. *SADNESS*
    # return *hint

    # Tuple subscripted by this PEP 646-compliant type variable tuple unpacked
    # into a PEP 646-compliant unpacked type variable tuple. This is insane,
    # because we're only going to rip this unpacked type variable tuple right
    # back out of this tuple. Blame the Python <= 3.10 interpreter.
    #
    # Note that a tuple was intentionally chosen for both space and time
    # efficiency. Although *ANY* container would satisfy Python <= 3.10, a tuple
    # has the advantage of being Python's most optimized container type.
    typevartuple_unpacked_parent = (*hint,)  # type: ignore[misc]

    # PEP 646-compliant unpacked child tuple hint subscripting this parent.
    typevartuple_unpacked = typevartuple_unpacked_parent[0]

    # Return this unpacked type variable tuple.
    return typevartuple_unpacked



#FIXME: Unit test us up, please. *sigh*
def make_hint_pep646_typevartuple_unpacked_subbed(
    hint: HintPep646TypeVarTupleType) -> Hint:
    '''
    Dynamically create and return a new :pep:`646`-compliant
    **subscription-based unpacked type variable tuple** (i.e., of the form
    ``typing.Unpack[hint]``) subscripted by the passed type variable tuple.

    Parameters
    ----------
    hint: HintPep646TypeVarTupleType
        Type variable tuple to be unpacked.

    Returns
    -------
    Hint
        Unpacked type variable tuple subscripted by this type variable tuple.
    '''

    # Avoid version-specific imports.
    #
    # Note that the PEP 646-compliant "typing.Unpack[...]" hint factory is
    # defined *ONLY* under Python >= 3.11. Thankfully, the PEP 646-compliant
    # "typing.TypeVarTuple" type is also defined *ONLY* under Python >= 3.11.
    # Since the caller passed an instance of that type, the active Python
    # interpreter *MUST* by definition target Python >= 3.11.
    from beartype.typing import Unpack  # pyright: ignore

    # If this hint is *NOT* a type variable tuple, raise an exception.
    die_unless_hint_pep646_typevartuple_packed(hint)  # pyright: ignore
    # Else, this hint is a type variable tuple.

    # Return this unpacked child tuple hint.
    return Unpack.__getitem__(hint)  # pyright: ignore



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **type-checking expression snippet class unit tests.**

This submodule unit tests the public API of the public
:mod:`beartype._check.code.snip.codesnipcls.PITH_INDEX_TO_VAR_NAME` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS                              }....................
def test_pith_index_to_var_name() -> None:
    '''
    Test the :obj:`beartype._check.code.snip.codesnipcls.PITH_INDEX_TO_VAR_NAME`
    dictionary singleton.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._check.code.snip.codesnipcls import PITH_INDEX_TO_VAR_NAME
    from beartype._data.code.datacodename import VAR_NAME_PITH_PREFIX
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this dictionary indexed by various positive integers creates and
    # returns the expected local pith variable names.
    assert PITH_INDEX_TO_VAR_NAME[1] == f'{VAR_NAME_PITH_PREFIX}1'
    assert PITH_INDEX_TO_VAR_NAME[2] == f'{VAR_NAME_PITH_PREFIX}2'

    # Assert this dictionary internally caches these constants.
    assert PITH_INDEX_TO_VAR_NAME[1] is PITH_INDEX_TO_VAR_NAME[1]
    assert PITH_INDEX_TO_VAR_NAME[2] is PITH_INDEX_TO_VAR_NAME[2]

    # ....................{ FAIL                           }....................
    # Assert that attempting to index this dictionary by non-integer indices
    # raises the expected exception.
    with raises(AssertionError):
        PITH_INDEX_TO_VAR_NAME[2.34]

    # Assert that attempting to index this dictionary by negative indices
    # raises the expected exception.
    with raises(AssertionError):
        PITH_INDEX_TO_VAR_NAME[-1]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **forward reference abstract base classes (ABCs)** (i.e., low-level
class hierarchy deferring the resolution of a stringified type hint referencing
an attribute that has yet to be defined and annotating a class or callable
decorated by the :func:`beartype.beartype` decorator).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintForwardRefException
from beartype.typing import (
    NoReturn,
    Optional,
    Type,
)
from beartype._data.typing.datatyping import (
    LexicalScope,
)
from beartype._check.forward.reference.fwdrefmeta import BeartypeForwardRefMeta

# ....................{ SUPERCLASSES                       }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: The names of *ALL* class variables declared below *MUST* be both:
# * Prefixed by "__beartype_".
# * Suffixed by "__".
#
# If this is *NOT* done, these variables could induce a namespace conflict with
# user-defined subpackages, submodules, and classes of the same names
# concatenated via the BeartypeForwardRefMeta.__getattr__() dunder method.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#FIXME: Unit test us up, please.
class BeartypeForwardRefABC(object, metaclass=BeartypeForwardRefMeta):
    '''
    Abstract base class (ABC) of all **forward reference subclasses** (i.e.,
    classes whose :class:`.BeartypeForwardRefMeta` metaclass defers the
    resolution of stringified type hints referencing actual type hints that have
    yet to be defined).

    Caveats
    -------
    **This ABC prohibits instantiation.** This ABC *only* exists to sanitize,
    simplify, and streamline the definition of subclasses passed as the second
    parameter to the :func:`isinstance` builtin, whose
    :class:`.BeartypeForwardRefMeta.__instancecheck__` dunder method then
    implicitly resolves the forward references encapsulated by those subclasses.
    The :func:`.make_forwardref_subtype` function dynamically creates and
    returns one concrete subclass of this ABC for each unique forward reference
    required by the :func:`beartype.beartype` decorator, whose :attr:`hint_name`
    class variable is the name of the attribute referenced by that reference.
    '''

    # ....................{ PRIVATE ~ class vars           }....................
    __name_beartype__: str = None  # type: ignore[assignment]
    '''
    Absolute (i.e., fully-qualified) or relative (i.e., unqualified) name of the
    type hint referenced by this forward reference subclass.
    '''


    __scope_name_beartype__: Optional[str] = None
    '''
    Fully-qualified name of the lexical scope to which the type hint referenced
    by this forward reference subclass is relative if that type hint is relative
    (i.e., if :attr:`__name_beartype__` is relative) *or* ignored otherwise
    (i.e., if :attr:`__name_beartype__` is absolute).
    '''

    # ....................{ INITIALIZERS                   }....................
    def __new__(cls, *args, **kwargs) -> NoReturn:
        '''
        Prohibit instantiation by unconditionally raising an exception.
        '''

        # Instantiatable. It's a word or my username isn't @UncleBobOnAStick.
        raise BeartypeDecorHintForwardRefException(
            f'{repr(BeartypeForwardRefABC)} subclass '
            f'{repr(cls)} not instantiatable.'
        )

    # ....................{ PRIVATE ~ testers              }....................
    @classmethod
    def __is_instance_beartype__(cls, obj: object) -> bool:
        '''
        :data:`True` only if the passed object is an instance of the external
        class referred to by this forward reference.

        Parameters
        ----------
        obj : object
            Arbitrary object to be tested.

        Returns
        -------
        bool
            :data:`True` only if this object is an instance of the external
            class referred to by this forward reference subclass.
        '''

        # # Resolve the external class referred to by this forward reference and
        # # permanently store that class in the "__type_beartype__" variable.
        # cls.__beartype_resolve_type__()

        # Return true only if this object is an instance of the external class
        # referenced by this forward reference.
        return isinstance(obj, cls.__type_beartype__)  # type: ignore[arg-type]


    @classmethod
    def __is_subclass_beartype__(cls, obj: object) -> bool:
        '''
        :data:`True` only if the passed object is a subclass of the external
        class referred to by this forward reference.

        Parameters
        ----------
        obj : object
            Arbitrary object to be tested.

        Returns
        -------
        bool
            :data:`True` only if this object is a subclass of the external class
            referred to by this forward reference subclass.
        '''

        # # Resolve the external class referred to by this forward reference and
        # # permanently store that class in the "__type_beartype__" variable.
        # cls.__beartype_resolve_type__()

        # Return true only if this object is a subclass of the external class
        # referenced by this forward reference.
        return issubclass(obj, cls.__type_beartype__)  # type: ignore[arg-type]

# ....................{ SUPERCLASSES ~ index               }....................
#FIXME: Unit test us up, please.
class _BeartypeForwardRefIndexedABC(BeartypeForwardRefABC):
    '''
    Abstract base class (ABC) of all **subscripted forward reference
    subclasses** (i.e., classes whose :class:`.BeartypeForwardRefMeta`
    metaclass defers the resolution of stringified type hints referencing actual
    type hints that have yet to be defined, subscripted by any arbitrary
    positional and keyword parameters).

    Subclasses of this ABC typically encapsulate user-defined generics that have
    yet to be declared (e.g., ``"MuhGeneric[int]"``).

    Caveats
    -------
    **This ABC currently ignores subscription.** Technically, this ABC *does*
    store all positional and keyword parameters subscripting this forward
    reference. Pragmatically, this ABC otherwise silently ignores these
    parameters by deferring to the superclass :meth:`.is_instance` method (which
    reduces to the trivial :func:`isinstance` call). Why? Because **generics**
    (i.e., :class:`typing.Generic` subclasses) themselves behave in the exact
    same way at runtime.
    '''

    # ....................{ PRIVATE ~ class vars           }....................
    __args_beartype__: tuple = None  # type: ignore[assignment]
    '''
    Tuple of all positional arguments subscripting this forward reference.
    '''


    __kwargs_beartype__: LexicalScope = None  # type: ignore[assignment]
    '''
    Dictionary of all keyword arguments subscripting this forward reference.
    '''


#FIXME: Unit test us up, please.
class _BeartypeForwardRefIndexableABC(BeartypeForwardRefABC):
    '''
    Abstract base class (ABC) of all **subscriptable forward reference
    subclasses** (i.e., classes whose :class:`.BeartypeForwardRefMeta`
    metaclass defers the resolution of stringified type hints referencing actual
    type hints that have yet to be defined, transparently permitting these type
    hints to be subscripted by any arbitrary positional and keyword parameters).
    '''

    # ....................{ DUNDERS                        }....................
    @classmethod
    def __class_getitem__(cls, *args, **kwargs) -> (
        Type[_BeartypeForwardRefIndexedABC]):
        '''
        Create and return a new **subscripted forward reference subclass**
        (i.e., concrete subclass of the :class:`._BeartypeForwardRefIndexedABC`
        abstract base class (ABC) deferring the resolution of the type hint with
        the passed name, subscripted by the passed positional and keyword
        arguments).

        This dunder method enables this forward reference subclass to
        transparently masquerade as any subscriptable type hint factory,
        including subscriptable user-defined generics that have yet to be
        declared (e.g., ``"MuhGeneric[int]"``).

        This dunder method is intentionally *not* memoized (e.g., by the
        :func:`callable_cached` decorator). Ideally, this dunder method *would*
        be memoized. Sadly, there exists no means of efficiently caching either
        non-variadic or variadic keyword arguments. Although technically
        feasible, doing so imposes practical costs defeating the entire point of
        memoization.
        '''

        # Avoid circular import dependencies.
        from beartype._check.forward.reference.fwdrefmake import (
            _make_forwardref_subtype)

        # Subscripted forward reference to be returned.
        forwardref_indexed_subtype: Type[_BeartypeForwardRefIndexedABC] = (
            _make_forwardref_subtype(  # type: ignore[assignment]
                hint_name=cls.__name_beartype__,
                scope_name=cls.__scope_name_beartype__,
                type_bases=_BeartypeForwardRefIndexedABC_BASES,
            ))

        # Classify the arguments subscripting this forward reference.
        forwardref_indexed_subtype.__args_beartype__ = args  # pyright: ignore[reportGeneralTypeIssues]
        forwardref_indexed_subtype.__kwargs_beartype__ = kwargs  # pyright: ignore[reportGeneralTypeIssues]

        # Return this subscripted forward reference.
        return forwardref_indexed_subtype

# ....................{ PRIVATE ~ tuples                   }....................
_BeartypeForwardRefIndexableABC_BASES = (_BeartypeForwardRefIndexableABC,)
'''
1-tuple containing *only* the :class:`._BeartypeForwardRefIndexableABC`
superclass to reduce space and time consumption.
'''


_BeartypeForwardRefIndexedABC_BASES = (_BeartypeForwardRefIndexedABC,)
'''
1-tuple containing *only* the :class:`._BeartypeForwardRefIndexedABC`
superclass to reduce space and time consumption.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype validator code snippets** (i.e., triple-quoted pure-Python code
constants formatted and concatenated together into wrapper functions
type-checking decorated callables annotated by one or more beartype
validators).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.code.datacodeindent import CODE_INDENT_1

# ....................{ INDENTATION                        }....................
VALE_CODE_INDENT_1 = f'{{indent}}{CODE_INDENT_1}'
'''
Code snippet prefixed by the placeholder substring ``"{indent}"`` (which the
:func:`beartype._check.code.codemain.make_func_pith_code` replaces with
the indentation level required by the current beartype validator) followed by a
single level of indentation.
'''

# ....................{ CHECK ~ factory                    }....................
VALE_CODE_CHECK_ISEQUAL_TEST = '''
{{indent}}# True only if this pith equals this object.
{{indent}}{{obj}} == {param_name_obj_value}'''
'''
:attr:`beartype.vale.IsEqual`-specific code snippet validating an arbitrary
object to be equal to another arbitrary object.
'''


VALE_CODE_CHECK_ISINSTANCE_TEST = '''
{{indent}}# True only if this pith is of this type.
{{indent}}isinstance({{obj}}, {param_name_types})'''
'''
:attr:`beartype.vale.IsInstance`-specific code snippet validating an arbitrary
object to instance an arbitrary type.
'''


VALE_CODE_CHECK_ISSUBCLASS_TEST = '''
{{indent}}# True only if this pith is a class subclassing this superclass.
{{indent}}(isinstance({{obj}}, type) and issubclass({{obj}}, {param_name_types}))'''
'''
:attr:`beartype.vale.IsSubclass`-specific code snippet validating an arbitrary
type to subclass another arbitrary type.
'''

# ....................{ CHECK ~ factory : isattr           }....................
VALE_CODE_CHECK_ISATTR_TEST = '''(
{{indent}}    # True only if this pith defines an attribute with this name.
{{indent}}    {attr_value_expr}
{{indent}}    is not {local_name_sentinel} and {attr_value_is_valid_expr}
{{indent}})'''
'''
:attr:`beartype.vale.IsAttr`-specific code snippet validating an arbitrary
object to define an attribute with an arbitrary name satisfying an arbitrary
expression evaluating to a boolean.
'''


_VALE_CODE_CHECK_ISATTR_VALUE_EXPR_RAW = (
    'getattr({{obj}}, {attr_name_expr}, {local_name_sentinel})')
'''
:attr:`beartype.vale.IsAttr`-specific Python expression inefficiently yielding
the value of the attribute with an arbitrary name of an arbitrary object to be
validated.
'''


VALE_CODE_CHECK_ISATTR_VALUE_EXPR = (
    f'({{local_name_attr_value}} := {_VALE_CODE_CHECK_ISATTR_VALUE_EXPR_RAW})')
'''
:attr:`beartype.vale.IsAttr`-specific Python expression efficiently yielding
the value of the attribute with an arbitrary name of an arbitrary object to be
validated.

For efficiency, this expression is optimized to localize this value to a local
variable whose name *must* be uniquified and formatted by the caller into the
``local_name_attr_value`` format variable.
'''

# ....................{ METHODS                            }....................
# Format methods of the code snippets declared above as a microoptimization.

VALE_CODE_CHECK_ISATTR_TEST_format = VALE_CODE_CHECK_ISATTR_TEST.format
VALE_CODE_CHECK_ISATTR_VALUE_EXPR_format = (
    VALE_CODE_CHECK_ISATTR_VALUE_EXPR.format)
VALE_CODE_CHECK_ISEQUAL_TEST_format = VALE_CODE_CHECK_ISEQUAL_TEST.format
VALE_CODE_CHECK_ISINSTANCE_TEST_format = VALE_CODE_CHECK_ISINSTANCE_TEST.format
VALE_CODE_CHECK_ISSUBCLASS_TEST_format = VALE_CODE_CHECK_ISSUBCLASS_TEST.format



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) procedural type
hint inference API** unit tests.

This submodule unit tests the public :mod:`beartype.bite.infer_hint` function.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_unless_package
from collections.abc import Iterable

# ....................{ TESTS                              }....................
def test_bite_infer_hint(
    bite_cases_infer_hint: (
    dict['beartype.BeartypeConf', Iterable[tuple[object, object]]])) -> None:
    '''
    Test the :func:`beartype.bite.infer_hint` function.

    Parameters
    ----------
    bite_cases_infer_hint : dict[beartype.BeartypeConf, Iterable[tuple[object, object]]]
        Dictionary mapping from each relevant beartype configuration to a
        corresponding iterable of all **type hint inference cases** specific to
        that configuration.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.bite import infer_hint
    from beartype.roar import BeartypeConfException
    from beartype._conf.confcommon import get_beartype_conf_strategy_on
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Default linear-time configuration.
    CONF_STRATEGY_ON = get_beartype_conf_strategy_on()

    # ....................{ PASS                           }....................
    # For each relevant beartype configuration...
    for conf, cases in bite_cases_infer_hint.items():
        # If this is the default linear-time configuration...
        if conf == CONF_STRATEGY_ON:
            # For each test object and expected type hint to be inferred from
            # that object...
            for obj, hint in cases:
                # Assert this function returns this type hint for this object
                # *WITHOUT* explicitly passing this configuration. Doing so
                # exercises that this function behaves as expected when passed
                # *NO* configuration.
                assert infer_hint(obj) == hint
        # Else, this is a non-default configuration. In this case...
        else:
            # For each test object and expected type hint to be inferred from
            # that object...
            for obj, hint in cases:
                # Assert this function returns this type hint for this object
                # under this non-default configuration.
                assert infer_hint(obj, conf=conf) == hint

    # ....................{ FAIL                           }....................
    # Assert that this function raises the expected exception when passed an
    # arbitrary object and an invalid beartype configuration.
    with raises(BeartypeConfException):
        infer_hint('Red, yellow,', conf='or ethereally pale,')


def test_bite_infer_hint_recursion() -> None:
    '''
    Test the :func:`beartype.bite.infer_hint` function with respect to
    **container recursion** (i.e., objects self-referentially containing
    themselves as items inside themselves).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import (
        BeartypeConf,
        BeartypeStrategy,
    )
    from beartype.bite import infer_hint
    from beartype.bite._infermain import BeartypeInferHintContainerRecursion
    from beartype.roar import BeartypeDoorInferHintRecursionWarning
    from pytest import warns

    # ..................{ LOCALS ~ list                      }..................
    # Recursive list containing *ONLY* a reference to itself and thus suitable
    # for O(1) constant-time type hint inference.
    LIST_RECURSIVE_STRATEGY_O1 = []
    LIST_RECURSIVE_STRATEGY_O1.append(LIST_RECURSIVE_STRATEGY_O1)

    # Recursive list containing a reference to itself as well as other unrelated
    # items and thus suitable for O(n) linear-time type hint inference.
    LIST_RECURSIVE_STRATEGY_ON = [
        "Rivals the pride of summer. 'Tis the haunt",
        b'Of every gentle wind, whose breath can teach',
    ]
    LIST_RECURSIVE_STRATEGY_ON.append(LIST_RECURSIVE_STRATEGY_ON)

    # ..................{ ASSERTS                            }..................
    # Assert that this function, when configured for O(1) constant-time type
    # hint inference and passed a recursive list containing *ONLY* a reference
    # to itself:
    # * Emits the expected warning.
    # * Returns the expected type hint.
    with warns(BeartypeDoorInferHintRecursionWarning):
        assert infer_hint(
            LIST_RECURSIVE_STRATEGY_O1,
            conf=BeartypeConf(strategy=BeartypeStrategy.O1),
        ) == list[BeartypeInferHintContainerRecursion]

    # Assert that this function, when configured for O(n) linear-time type hint
    # inference and passed a recursive list containing a reference to itself as
    # well as other unrelated itemsf:
    # * Emits the expected warning.
    # * Returns the expected type hint.
    with warns(BeartypeDoorInferHintRecursionWarning):
        assert infer_hint(LIST_RECURSIVE_STRATEGY_ON) == list[
            str | bytes | BeartypeInferHintContainerRecursion ]

# ....................{ TESTS ~ third-party                }....................
# Tests exercising infer_hint() with respect to third-party packages.

def test_bite_infer_hint_numpy(
    bite_cases_infer_hint: 'Iterable[Tuple[object, object]]',
    numpy_arrays: 'beartype_test.a00_unit.data.api.data_apinumpy._NumpyArrays',
) -> None:
    '''
    Test the :func:`beartype.bite.infer_hint` function with respect to
    third-party :mod:`numpy` arrays if :mod:`numpy` is importable *or* silently
    skip this test otherwise (i.e., if :mod:`numpy` is unimportable).

    Parameters
    ----------
    bite_cases_infer_hint : Iterable[Tuple[object, object]]
        Iterable of **type hint inference cases** (i.e., 2-tuples ``(obj,
        hint)`` describing the type hint matching an arbitrary object).
    numpy_arrays : beartype_test.a00_unit.data.api.data_apinumpy._NumpyArrays
        **NumPy arrays dataclass** (i.e., well-typed and -described object
        comprising various :mod:`numpy` arrays).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.bite import infer_hint
    from beartype.vale import (
        IsAttr,
        IsEqual,
    )
    from beartype._util.api.standard.utiltyping import import_typing_attr_or_none
    from numpy import unsignedinteger
    from numpy.typing import NDArray

    # ....................{ LOCALS                         }....................
    # PEP 593-compliant "Annotated" type hint factory imported from either the
    # standard "typing" or third-party "typing_extensions" modules if importable
    # from at least one of those modules *OR* "None" otherwise.
    Annotated = import_typing_attr_or_none('Annotated')

    # List of all NumPy type hint inference cases (i.e., 2-tuples "(obj, hint)"
    # describing the type hint matching a NumPy array).
    INFER_HINT_NUMPY_CASES = [
        # ..................{ NUMPY                          }..................
        # One-dimensional NumPy boolean arrays are annotated as either...
        (numpy_arrays.array_1d_boolean, (
            # If "typing(|_extensions).Annotated" is importable, the third-party
            # "numpy.typing.NDArray" type hint factory subscripted by the
            # builtin "bool" type annotated to require a one-dimensional array.
            Annotated[NDArray[bool], IsAttr['ndim', IsEqual[1]]]
            if Annotated is not None else
            # Else, "typing(|_extensions).Annotated" is unimportable. In this
            # case, merely this subscripted factory.
            NDArray[bool]
        )),
        # One-dimensional NumPy complex arrays are annotated as either...
        (numpy_arrays.array_1d_complex_128, (
            # If "typing(|_extensions).Annotated" is importable, the third-party
            # "numpy.typing.NDArray" type hint factory subscripted by the
            # builtin "complex" type annotated to require a one-dimensional
            # array.
            Annotated[NDArray[complex], IsAttr['ndim', IsEqual[1]]]
            if Annotated is not None else
            # Else, "typing(|_extensions).Annotated" is unimportable. In this
            # case, merely this subscripted factory.
            NDArray[complex]
        )),
        # One-dimensional NumPy 32-bit floating point arrays are annotated as
        # either...
        (numpy_arrays.array_1d_float_32, (
            # If "typing(|_extensions).Annotated" is importable, the third-party
            # "numpy.typing.NDArray" type hint factory subscripted by the
            # builtin "float" type annotated to require a one-dimensional array.
            Annotated[NDArray[float], IsAttr['ndim', IsEqual[1]]]
            if Annotated is not None else
            # Else, "typing(|_extensions).Annotated" is unimportable. In this
            # case, merely this subscripted factory.
            NDArray[float]
        )),
        # One-dimensional NumPy 64-bit floating point arrays are annotated as
        # either...
        (numpy_arrays.array_1d_float_64, (
            # If "typing(|_extensions).Annotated" is importable, the third-party
            # "numpy.typing.NDArray" type hint factory subscripted by the
            # builtin "float" type annotated to require a one-dimensional array.
            Annotated[NDArray[float], IsAttr['ndim', IsEqual[1]]]
            if Annotated is not None else
            # Else, "typing(|_extensions).Annotated" is unimportable. In this
            # case, merely this subscripted factory.
            NDArray[float]
        )),
        # One-dimensional NumPy 32-bit integer arrays are annotated as either...
        (numpy_arrays.array_1d_int_32, (
            # If "typing(|_extensions).Annotated" is importable, the third-party
            # "numpy.typing.NDArray" type hint factory subscripted by the
            # builtin "int" type annotated to require a one-dimensional array.
            Annotated[NDArray[int], IsAttr['ndim', IsEqual[1]]]
            if Annotated is not None else
            # Else, "typing(|_extensions).Annotated" is unimportable. In this
            # case, merely this subscripted factory.
            NDArray[int]
        )),
        # One-dimensional NumPy 64-bit integer arrays are annotated as either...
        (numpy_arrays.array_1d_int_64, (
            # If "typing(|_extensions).Annotated" is importable, the third-party
            # "numpy.typing.NDArray" type hint factory subscripted by the
            # builtin "int" type annotated to require a one-dimensional array.
            Annotated[NDArray[int], IsAttr['ndim', IsEqual[1]]]
            if Annotated is not None else
            # Else, "typing(|_extensions).Annotated" is unimportable. In this
            # case, merely this subscripted factory.
            NDArray[int]
        )),
        # One-dimensional NumPy 32-bit unsigned integer arrays are annotated as
        # either...
        (numpy_arrays.array_1d_uint_32, (
            # If "typing(|_extensions).Annotated" is importable, the third-party
            # "numpy.typing.NDArray" type hint factory subscripted by the
            # "numpy.unsignedinteger" dtype annotated to require a
            # one-dimensional array.
            Annotated[NDArray[unsignedinteger], IsAttr['ndim', IsEqual[1]]]
            if Annotated is not None else
            # Else, "typing(|_extensions).Annotated" is unimportable. In this
            # case, merely this subscripted factory.
            NDArray[unsignedinteger]
        )),
        # One-dimensional NumPy 64-bit unsigned integer arrays are annotated as
        # either...
        (numpy_arrays.array_1d_uint_64, (
            # If "typing(|_extensions).Annotated" is importable, the third-party
            # "numpy.typing.NDArray" type hint factory subscripted by the
            # "numpy.unsignedinteger" dtype annotated to require a
            # one-dimensional array.
            Annotated[NDArray[unsignedinteger], IsAttr['ndim', IsEqual[1]]]
            if Annotated is not None else
            # Else, "typing(|_extensions).Annotated" is unimportable. In this
            # case, merely this subscripted factory.
            NDArray[unsignedinteger]
        )),
        # One-dimensional NumPy array serving as a memory view over arbitrary
        # bytes all of the same length are annotated as either...
        (numpy_arrays.array_1d_memory_view, (
            # If "typing(|_extensions).Annotated" is importable, the third-party
            # "numpy.typing.NDArray" type hint factory subscripted by the
            # builtin "memoryview" type annotated to require a one-dimensional
            # array.
            Annotated[NDArray[memoryview], IsAttr['ndim', IsEqual[1]]]
            if Annotated is not None else
            # Else, "typing(|_extensions).Annotated" is unimportable. In this
            # case, merely this subscripted factory.
            NDArray[memoryview]
        )),
        # One-dimensional NumPy byte string arrays are annotated as either...
        (numpy_arrays.array_1d_string_byte, (
            # If "typing(|_extensions).Annotated" is importable, the third-party
            # "numpy.typing.NDArray" type hint factory subscripted by the
            # builtin "bytes" type annotated to require a one-dimensional array.
            Annotated[NDArray[bytes], IsAttr['ndim', IsEqual[1]]]
            if Annotated is not None else
            # Else, "typing(|_extensions).Annotated" is unimportable. In this
            # case, merely this subscripted factory.
            NDArray[bytes]
        )),
        # One-dimensional NumPy Unicode string arrays are annotated as either...
        (numpy_arrays.array_1d_string_char, (
            # If "typing(|_extensions).Annotated" is importable, the third-party
            # "numpy.typing.NDArray" type hint factory subscripted by the
            # builtin "str" type annotated to require a one-dimensional array.
            Annotated[NDArray[str], IsAttr['ndim', IsEqual[1]]]
            if Annotated is not None else
            # Else, "typing(|_extensions).Annotated" is unimportable. In this
            # case, merely this subscripted factory.
            NDArray[str]
        )),
    ]

    # ....................{ ASSERTS                        }....................
    # For each type hint inference case to be tested...
    for obj, hint in INFER_HINT_NUMPY_CASES:
        # Assert this function returns the expected type hint for this object.
        assert infer_hint(obj) == hint


@skip_unless_package('pygments')
def test_bite_infer_hint_pygments() -> None:
    '''
    Test the :func:`beartype.bite.infer_hint` function with respect to
    third-party :mod:`pygments` package if that package is importable *or*
    silently skip this test otherwise (i.e., if that package is unimportable).
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.bite import infer_hint
    from beartype._data.hint.sign.datahintsigns import HintSignList
    from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none
    from pygments.lexers import PythonLexer

    # ....................{ LOCALS                         }....................
    # List of all root tokens exposed by the pygments Python lexer, exercising
    # edge cases in the infer_hint() function that continued to break. Ugh!
    root_tokens = PythonLexer().tokens['root']

    # Hint inferred for this list, implicitly exercising that the infer_hint()
    # function successfully infers this hint *WITHOUT* raising an exception --
    # which it used to.
    root_tokens_hint = infer_hint(root_tokens)

    # ....................{ ASSERTS                        }....................
    # Assert that this hint is a subscripted "list[...]" hint.
    #
    # Note that we intentionally do *NOT* assert the specific child hints
    # subscripting this hint. Why? Because these child hints include references
    # to private "pygments" types (e.g., "pygments._token.TokenType") dependent
    # on the current pygments version. For that reason, we intentionally omit
    # this hint from the "bite_cases_infer_hint" fixture leveraged above.
    assert get_hint_pep_sign_or_none(root_tokens_hint) is HintSignList



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484` and :pep:`585` **generic type hint getter** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget`
submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# from beartype_test._util.mark.pytskip import skip_if_python_version_less_than

# ....................{ TESTS ~ args                       }....................
def test_get_hint_pep484585_generic_args_full() -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484585.pep484585generic.get_hint_pep484585_generic_args_full`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep484585Exception
    from beartype.typing import (
        Generic,
        List,
        Sequence,
    )
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
        get_hint_pep484585_generic_args_full)
    from beartype_test.a00_unit.data.pep.generic.data_pep484generic import (
        Nongeneric,
        Pep484GenericST,
    )
    from beartype_test.a00_unit.data.pep.generic.data_pep585generic import (
        Pep585SequenceU,
    )
    from beartype_test.a00_unit.data.pep.generic.data_pep484585generic import (
        Pep484585GenericSTSequenceU,
        Pep484585GenericIntTSequenceU,
        Pep484585GenericUUST,
        Pep484585GenericUIntT,
    )
    from beartype_test.a00_unit.data.pep.data_pep484 import (
        S,
        T,
        U,
    )
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # List of all generic argument cases, each of which is a 2-tuple of the
    # form "(src_generic, trg_args)" such that:
    # * "src_generic" is a PEP 484- or 585-compliant generic to be passed as the
    #   input "hint" parameter to this getter.
    # * "trg_args" is the output tuple returned by this getter when passed that
    #   input generic.
    PEP484585_GENERIC_ARGS_FULL = [
        # ....................{ PEP ~ 484                  }....................
        (Pep484GenericST, (S, T,)),
        (Pep484GenericST[int, float], (int, float,)),
        (Pep585SequenceU, (U,)),
        (Pep585SequenceU[complex], (complex,)),
        (Pep484585GenericSTSequenceU, (bool, int, T, U,)),
        (Pep484585GenericIntTSequenceU, (bool, int, float, U,)),
        (Pep484585GenericUUST, (U, S, T, U,)),

        # ....................{ PEP ~ 585                  }....................
        (Pep484585GenericSTSequenceU[float, complex], (bool, int, float, complex,)),
        (Pep484585GenericIntTSequenceU[complex], (bool, int, float, complex,)),
        (Pep484585GenericUUST[bool, int, float], (bool, int, float, bool,)),
        (Pep484585GenericUIntT, (U, int, T, U,)),
        (Pep484585GenericUIntT[bool, float], (bool, int, float, bool,)),
    ]

    # List of all generic argument cases, each of which is a 2-tuple of the
    # form "(src_generic, src_base_target, trg_args)" such that:
    # * "src_generic" is a PEP 484- or 585-compliant generic to be passed as the
    #   input "hint" parameter to this getter.
    # * "src_base_target" is a target pseudo-superclass to be passed as the
    #   input "hint_base_target" parameter to this getter.
    # * "trg_args" is the output tuple returned by this getter when passed that
    #   input generic and target pseudo-superclass.
    PEP484585_GENERIC_BASE_TARGET_ARGS_FULL = [
        # ....................{ PEP ~ 484                  }....................
        (Pep484GenericST, Generic, (S, T,)),
        (Pep484GenericST, Generic[S, T], (S, T,)),
        (Pep484GenericST[int, float], Generic, (int, float,)),
        (Pep585SequenceU, Sequence, (U,)),
        (Pep585SequenceU[complex], Sequence, (complex,)),
        (Pep484585GenericSTSequenceU, List, (bool,)),
        (Pep484585GenericSTSequenceU, Pep484GenericST, (int, T,)),
        (Pep484585GenericSTSequenceU, Nongeneric, ()),
        (Pep484585GenericSTSequenceU, Pep585SequenceU, (U,)),
        (Pep484585GenericIntTSequenceU, Pep484585GenericSTSequenceU, (bool, int, float, U,)),
        (Pep484585GenericUUST, Pep585SequenceU, (U,)),
        (Pep484585GenericUUST, Pep484GenericST, (S, T)),

        # ....................{ PEP ~ 585                  }....................
        (Pep484585GenericSTSequenceU[float, complex], Pep484GenericST, (int, float,)),
        (Pep484585GenericSTSequenceU[float, complex], Pep585SequenceU, (complex,)),
        (Pep484585GenericIntTSequenceU[complex], Pep484585GenericSTSequenceU, (bool, int, float, complex,)),
        (Pep484585GenericUUST[bool, int, float], List[U], (bool,)),
        (Pep484585GenericUUST[bool, int, float], Pep484GenericST, (int, float)),
        (Pep484585GenericUIntT, Pep585SequenceU, (U,)),
        (Pep484585GenericUIntT, Pep484GenericST, (int, T)),
        (Pep484585GenericUIntT[bool, float], List[U], (bool,)),
        (Pep484585GenericUIntT[bool, float], Pep484GenericST, (int, float)),
    ]

    # ....................{ PASS                           }....................
    # Assert that this getter returns the expected tuples for the passed
    # generics, including subscripted and unsubscripted forms of these generics.
    for src_generic, trg_args in PEP484585_GENERIC_ARGS_FULL:
        assert get_hint_pep484585_generic_args_full(src_generic) == trg_args

        # Additionally assert that this getter returns the same tuple when
        # passed the passed generic as the target pseudo-superclass. By
        # definition, any class is the superclass of itself. (Math goes hard.)
        assert get_hint_pep484585_generic_args_full(
            src_generic, src_generic) == trg_args

    # Assert that this getter returns the expected tuples for the passed
    # generics and target pseudo-superclasses of these generics, including
    # subscripted and unsubscripted forms of these generics.
    for src_generic, src_base_target, trg_args in (
        PEP484585_GENERIC_BASE_TARGET_ARGS_FULL):
        assert get_hint_pep484585_generic_args_full(
            src_generic, src_base_target) == trg_args

    # ....................{ FAIL                           }....................
    # Assert that this getter raises the expected exception when passed an
    # object that is *NOT* a PEP 484- or 585-compliant generic.
    with raises(BeartypeDecorHintPep484585Exception):
        get_hint_pep484585_generic_args_full(Nongeneric)

# ....................{ TESTS ~ base                       }....................
def test_get_hint_pep484585_generic_base_extrinsic_sign_or_none() -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget.get_hint_pep484585_generic_base_extrinsic_sign_or_none`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.typing import (
        Generic,
        NamedTuple,
        TypedDict,
    )
    from beartype._data.typing.datatyping import T
    from beartype._data.hint.sign.datahintsigns import (
        HintSignNamedTuple,
        HintSignTypedDict,
    )
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
        get_hint_pep484585_generic_base_extrinsic_sign_or_none)

    # ....................{ CLASSES                        }....................
    #FIXME: Useful docstring for when we use the generics defined below.
    # Python < 3.11 fails to support multiple inheritance in concert with the
    # :class:`typing.Generic` superclass and either of the
    # :class:`typing.NamedTuple` or :class:`typing.TypedDict` superclasses. Ergo,
    # Python < 3.11 fails to support either generic named tuples *or* generic
    # typed dictionaries. Ergo, Python < 3.11 fails to support this unit test.

    #FIXME: When we get around to supporting this, note that "typing.NamedTuple"
    #*ONLY* supports multiple inheritance in partnership with "typing.Generic"
    #under Python >= 3.11. Under older Python versions, the "typing" module
    #raises the following exception on attempting to define such subclasses:
    #    TypeError: Multiple inheritance with NamedTuple is not supported
    # class GenericNamedTuple(NamedTuple, Generic[T]):
    #     '''
    #     Arbitrary **generic named tuple** (i.e., type subclassing both the
    #     :pep:`484`-compliant :class:`typing.Generic` superclass *and* the
    #     :pep:`484`-compliant :class:`typing.NamedTuple` superclass).
    #     '''
    #
    #     generic_item: T


    #FIXME: Nice, but currently unused.
    # class GenericTypedDict(TypedDict, Generic[T]):
    #     '''
    #     Arbitrary **generic typed dictionary** (i.e., type subclassing both the
    #     :pep:`484`-compliant :class:`typing.Generic` superclass *and* the
    #     :pep:`589`-compliant :class:`typing.TypedDict` superclass).
    #     '''
    #
    #     generic_item: T

    # ....................{ LOCALS                         }....................
    # Tuple of 2-tuples "(hint_base, hint_extrinsic_sign)" where:
    # * "hint_base" is a valid pseudo-superclass of a PEP 484- or 585-compliant
    #   generic.
    # * "hint_extrinsic_sign" is either:
    #   * If that pseudo-superclass is extrinsic, the sign uniquely identifying
    #     that pseudo-superclass.
    #   * If that pseudo-superclass is intrinsic, "None".
    TEST_GENERIC_BASES_EXTRINSIC_SIGNS = (
        # Intrinsic PEP 484-compliant "typing.Generic" pseudo-superclass in both
        # subscripted and unsubscripted forms.
        (Generic, None),
        (Generic[T], None),

        # Extrinsic PEP 484-compliant "typing.NamedTuple" pseudo-superclass.
        (NamedTuple, HintSignNamedTuple),

        # Extrinsic PEP 589-compliant "typing.TypedDict" pseudo-superclass.
        (TypedDict, HintSignTypedDict),
    )

    # ....................{ ASSERTS                        }....................
    # For each such unsubscripted generic and additional non-generic sign...
    for hint_base, hint_extrinsic_sign in TEST_GENERIC_BASES_EXTRINSIC_SIGNS:
        # Assert that this getter passed this generic returns this sign.
        assert get_hint_pep484585_generic_base_extrinsic_sign_or_none(
            hint_base) is hint_extrinsic_sign


def test_get_hint_pep484585_generic_bases_unerased(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget.get_hint_pep484585_generic_bases_unerased`
    getter.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of PEP-compliant type hint metadata describing sample PEP-compliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPepException
    from beartype._data.hint.sign.datahintsignset import HINT_SIGNS_GENERIC
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
        get_hint_pep484585_generic_bases_unerased)
    from beartype._util.hint.pep.utilpeptest import is_hint_pep_type_typing
    from beartype_test.a00_unit.data.hint.data_hint import NOT_HINTS_PEP
    from pytest import raises

    # For each PEP-compliant hint to be tested...
    for hint_pep_meta in hints_pep_meta:
        # If this is a generic hint, assert that this getter returns the
        # expected unerased pseudo-superclasses of this generic.
        if hint_pep_meta.pep_sign in HINT_SIGNS_GENERIC:
            hint_pep_bases = get_hint_pep484585_generic_bases_unerased(
                hint_pep_meta.hint)
            assert isinstance(hint_pep_bases, tuple)
            assert hint_pep_bases
        # Else, this is *NOT* a generic hint.
        #
        # If this hint is *NOT* defined by the "typing" module, assert that this
        # getter raises the expected exception.
        elif not is_hint_pep_type_typing(hint_pep_meta.hint):
            # print(f'hint_pep_meta: {repr(hint_pep_meta)}')
            with raises(BeartypeDecorHintPepException):
                get_hint_pep484585_generic_bases_unerased(hint_pep_meta.hint)
        # Else, this hint is defined by the "typing" module. In this case, this
        # hint may or may not be implemented as a generic conditionally
        # depending on the current Python version -- especially under the
        # Python < 3.7.0 implementations of the "typing" module, where
        # effectively *EVERYTHING* was internally implemented as a generic.
        # While we could technically correct for this conditionality, doing so
        # would render the resulting code less maintainable for no useful gain.
        # Ergo, we quietly ignore this edge case and get on with actual coding.

    # Assert this getter raises the expected exception for non-PEP-compliant
    # type hints.
    for not_hint_pep in NOT_HINTS_PEP:
        with raises(BeartypeDecorHintPepException):
            assert get_hint_pep484585_generic_bases_unerased(not_hint_pep)


def test_get_hint_pep484585_generic_base_in_module_first() -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget.get_hint_pep484585_generic_base_in_module_first`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep484585Exception
    from beartype.typing import Generic
    from beartype._data.typing.datatyping import T
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
        get_hint_pep484585_generic_base_in_module_first)
    from pytest import raises

    # ....................{ CLASSES                        }....................
    class ToWhichThisClosingNight(object):
        '''
        Arbitrary superclass.
        '''

        pass


    class OfTheDyingYear(Generic[T], str, ToWhichThisClosingNight):
        '''
        Arbitrary generic subclassing:

        * The generic superclass subscripted by a type variable.
        * An arbitrary type *not* defined by the :mod:`beartype_test` package.
        * An arbitrary type defined by the :mod:`beartype_test` package.
        '''

        pass

    # ....................{ LOCALS                         }....................
    # Tuple of all generic type hints to be tested below, including...
    TEST_GENERICS = (
        # An unsubscripted generic.
        OfTheDyingYear,
        # A subscripted generic (subscripted by an arbitrary type).
        OfTheDyingYear[int],
    )

    # ....................{ ASSERTS                        }....................
    # For each such generic type hint...
    for test_generic in TEST_GENERICS:
        # ....................{ PASS                       }....................
        # Assert that this finder passed this generic returns the first unerased
        # superclass of this generic declared by this package.
        assert get_hint_pep484585_generic_base_in_module_first(
            hint=test_generic, module_names=frozenset(('beartype_test',))) is (
            ToWhichThisClosingNight)

        # ....................{ FAIL                       }....................
        # Assert that this finder passed this generic and the name of a
        # hypothetical module guaranteed *NOT* to exist raises the expected
        # exception.
        with raises(BeartypeDecorHintPep484585Exception):
            get_hint_pep484585_generic_base_in_module_first(
                hint=test_generic,
                module_names=frozenset((
                    'will_be.the.dome_of.a_vast_sepulchre',)),
            )

# ....................{ TESTS ~ type                       }....................
def test_get_hint_pep484585_generic_type_or_none(hints_pep_meta) -> None:
    '''
    Test the
    :func:`beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget.get_hint_pep484585_generic_type_or_none`
    getter.

    Parameters
    ----------
    hints_pep_meta : List[beartype_test.a00_unit.data.hint.util.data_hintmetacls.HintPepMetadata]
        List of PEP-compliant type hint metadata describing sample PEP-compliant
        type hints exercising edge cases in the :mod:`beartype` codebase.
    '''

    # Defer test-specific imports.
    from beartype._data.hint.sign.datahintsigns import HintSignPep484585GenericUnsubbed
    from beartype._util.hint.pep.proposal.pep484585.generic.pep484585genget import (
        get_hint_pep484585_generic_type_or_none)

    # Assert this getter returns the expected type origin for all
    # PEP-compliant type hint generics. While we could support non-generics as
    # well, there's little benefit and significant costs to doing so. Instead,
    # we assert this getter only returns the expected type origin for a small
    # subset of type hints.
    for hint_pep_meta in hints_pep_meta:
        if hint_pep_meta.pep_sign is HintSignPep484585GenericUnsubbed:
            assert get_hint_pep484585_generic_type_or_none(
                hint_pep_meta.hint) is hint_pep_meta.generic_type

    #FIXME: Uncomment if we ever want to exercise extreme edge cases. *shrug*
    # from beartype_test.a00_unit.data.hint.data_hint import NOT_HINTS_PEP
    #
    # # Assert this getter returns the expected type origin for all
    # # PEP-compliant type hints.
    # for hint_pep_meta in HINTS_PEP_META:
    #     assert get_hint_pep484585_generic_type_or_none(
    #         hint_pep_meta.hint) is hint_pep_meta.generic_type
    #
    # # Assert this getter raises the expected exception for non-PEP-compliant
    # # type hints.
    # for not_hint_pep in NOT_HINTS_PEP:
    #     assert get_hint_pep484585_generic_type_or_none(not_hint_pep) is None



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator** :pep:`646`-compliant :obj:`typing.Self` **unit tests**.

This submodule unit tests :pep:`646` support implemented in the
:func:`beartype.beartype` decorator.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype_test._util.mark.pytskip import skip_if_python_version_less_than

# ....................{ TESTS                              }....................
@skip_if_python_version_less_than('3.11.0')
def test_decor_pep646() -> None:
    '''
    Test :pep:`646` support implemented in the :func:`beartype.beartype`
    decorator if the active Python interpreter targets Python >= 3.11 *or* skip
    otherwise.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype import beartype
    from beartype.roar import BeartypeDecorHintPep646692Exception
    from beartype.typing import (
        Generic,
        Tuple,
        TypeVarTuple,
        Unpack,
    )
    from beartype._util.hint.pep.proposal.pep646692 import (
        make_hint_pep646_typevartuple_unpacked_prefix,
        make_hint_pep646_typevartuple_unpacked_subbed,
    )
    from pytest import raises

    # ....................{ LOCALs                         }....................
    # Arbitrary type variable tuple.
    Ts = TypeVarTuple('Ts')

    # Prefix-based unpacked type variable tuple of the form "*Ts".
    Ts_prefix = make_hint_pep646_typevartuple_unpacked_prefix(Ts)

    # Subscription-based unpacked type variable tuple of the form
    # "typing.Unpack[Ts]".
    Ts_subbed = make_hint_pep646_typevartuple_unpacked_subbed(Ts)

    # ....................{ CLASSES                        }....................
    @beartype
    class ItsMotions(Generic[Ts_subbed]):
        '''
        Generic over an arbitrary number of type variables, decorated by
        :func:`beartype.beartype` and defined by subclassing the
        :class:`.Generic` superclass subscripted by the :pep:`646`-compliant
        :obj:`typing.Unpack` type hint factory subscripted by a
        :pep:`646`-compliant type variable tuple.

        Note that this is semantically equivalent to:

        .. code-block:: python

           class ItsMotions(Generic[*Ts]):
        '''

        pass


    # Arbitrary instance of this class.
    render_up_its_majesty = ItsMotions()

    # ....................{ CALLABLES                      }....................
    @beartype
    def of_all_the_grace(
        and_to_the_damp_leaves: ItsMotions,
        *and_beauty_that_endued: Ts_prefix
    ) -> Tuple[ItsMotions, Ts_prefix]:
        '''
        Arbitrary callable simply returning the tuple of all passed variadic
        positional parameters prepended by the passed generic, decorated by
        :func:`beartype.beartype` and correctly annotated by:

        * A variadic positional parameter typed as a :pep:`646`-compliant
          unpacked type variable tuple hint.
        * A return typed as a :pep:`484`-compliant tuple type hint subscripted
          by the same :pep:`646`-compliant unpacked type variable tuple hint.
        '''

        # Return this variadic positional parameter prepended by this generic.
        return (and_to_the_damp_leaves,) + and_beauty_that_endued

    # ....................{ PASS                           }....................
    # Assert that this callable returns a tuple of all passed positional
    # parameters.
    assert of_all_the_grace(
        render_up_its_majesty, 'Of all the grace', 'and beauty', 'that endued') == (
        render_up_its_majesty, 'Of all the grace', 'and beauty', 'that endued')

    # ....................{ FAIL                           }....................
    # Assert that @beartype raises the expected exception when decorating a
    # callable annotated by a hint unpacking a PEP 646-noncompliant object
    # (i.e., *ANY* object other than a PEP 646-compliant type variable tuple).
    with raises(BeartypeDecorHintPep646692Exception):
        @beartype
        def scatter_its_music(
            *on_the_unfeeling_storm: Unpack[Tuple[str]]) -> str:
            return on_the_unfeeling_storm[0]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **type-checking expression snippet classes** (i.e., low-level classes
dynamically and efficiently generating substrings intended to be interpolated
into boolean expressions type-checking arbitrary objects against various type
hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.code.datacodename import VAR_NAME_PITH_PREFIX

# ....................{ PRIVATE ~ subclasses               }....................
class _HintIndexToHintPlaceholder(dict):
    '''
    **Local type-checking placeholder cache** (i.e., dictionary mapping from the
    1-based index uniquely identifying each **pith** (i.e., current parameter or
    return value *or* item contained in the current parameter or return value
    type-checked by the current call in the body of a runtime type-checker
    dynamically generated by :mod:`beartype`) to the corresponding type-checking
    placeholder substring).

    Design
    ------
    Each value mapped by this cache is a **type-checking placeholder substring**
    to be globally replaced in the **type-checking wrapper function code
    snippet** (i.e., the ``func_wrapper_code`` local defined by the
    :func:`beartype._check.code.codemain.make_check_expr` factory) by a Python
    code snippet type-checking the **current pith expression** (i.e., the
    ``pith_var_name`` local) against the **currently visited type hint** (i.e.,
    the :attr:`hint` instance variable).

    Each such substring provides indirection enabling the currently visited
    parent hint to defer and delegate the generation of code type-checking each
    child type hint of this parent hint to the subsequent time at which that
    child type hint is visited by that BFS.

    Each such substring is intentionally prefixed and suffixed by characters
    that:

    * Are intentionally invalid as Python code, guaranteeing that the top-level
      call to the func:`exec` builtin subsequently performed by the
      :func:`beartype.beartype` decorator will raise a :exc:`SyntaxError`
      exception if the caller failed to replace *all* placeholder substrings.
    * Protect the :attr:`pith_var_name_index` integer embedded in this substring
      against ambiguous global replacements of longer such integers containing
      this integer (e.g., the integer 1, contained in the longer integers 11 and
      21). If this integer were *not* protected in this manner, then the first
      substring ``"0"`` would ambiguously overlap with the subsequent
      substring ``"10"``, which would then produce catastrophically erroneous
      and undebuggable Python code.

    Example
    -------
    For example, the :func:`beartype._check.code.codemain.make_check_expr`
    factory might generate intermediary code resembling the following on
    visiting the :obj:`typing.Union` parent type hint of a subscripted type hint
    ``typing.Union[int, str]`` *before* visiting either the :class:`int` or
    :class:`str` child type hints of that parent type hint:

    .. code-block:: python

       if not (
           @{0}! or
           @{1}!
       ):
           raise get_func_pith_violation(
               func=__beartype_func,
               pith_name=$%PITH_ROOT_NAME/~,
               pith_value=__beartype_pith_root,
           )

    Note the unique substrings ``"@{0}!"`` and ``"@{1}!"`` in that code, which
    that factory iteratively replaces with code type-checking each of the child
    type hints (e.g., :class:`int`, :class:`str`) subscripting that
    :obj:`typing.Union` parent type hint. The final code memoized by that
    factory might then resemble:

    .. code-block:: python

       if not (
           isinstance(__beartype_pith_root, int) or
           isinstance(__beartype_pith_root, str)
       ):
           raise get_func_pith_violation(
               func=__beartype_func,
               pith_name=$%PITH_ROOT_NAME/~,
               pith_value=__beartype_pith_root,
           )

    See Also
    --------
    :data:`.PITH_INDEX_TO_HINT_PLACEHOLDER`
        Singleton instance of this dictionary subclass.
    '''

    # ....................{ DUNDERS                        }....................
    def __missing__(self, pith_index: int) -> str:
        '''
        Dunder method explicitly called by the superclass
        :meth:`dict.__getitem__` method implicitly called on the first ``[``-
        and ``]``-delimited attempt to access a local type-checking placeholder
        uniquely identified by the passed 1-based index.

        Parameters
        ----------
        pith_index : int
            1-based index suffixing the local type-checking placeholder to be
            created, cached, and returned.

        Returns
        -------
        str
            Prospective type-checking placeholder of this local pith variable.

        Raises
        ------
        AssertionError
            If either:

            * ``pith_index`` is *not* an integer.
            * ``pith_index`` is a **negative integer** (i.e., less than 0).
        '''
        assert isinstance(pith_index, int), f'{repr(pith_index)} not integer.'
        assert pith_index >= 0, f'{pith_index} < 0.'
        # print(f'Generating indentation level {indent_level}...')

        # Placeholder substring to be globally replaced by code type-checking
        # the current pith against this hint.
        hint_placeholder = (
            f'{_HINT_PLACEHOLDER_PREFIX}{pith_index}'
            f'{_HINT_PLACEHOLDER_SUFFIX}'
        )

        # Cache this placeholder substring.
        self[pith_index] = hint_placeholder

        # Return this placeholder substring.
        return hint_placeholder


class _PithIndexToVarName(dict):
    '''
    **Local pith variable name cache** (i.e., dictionary mapping from the
    1-based index uniquely identifying each **pith** (i.e., current parameter or
    return value *or* item contained in the current parameter or return value
    type-checked by the current call in the body of a runtime type-checker
    dynamically generated by :mod:`beartype`) to the corresponding name of a
    prospective local variable assigned that value in that body).

    See Also
    --------
    :data:`.PITH_INDEX_TO_VAR_NAME`
        Singleton instance of this dictionary subclass.
    '''

    # ....................{ DUNDERS                        }....................
    def __missing__(self, pith_index: int) -> str:
        '''
        Dunder method explicitly called by the superclass
        :meth:`dict.__getitem__` method implicitly called on the first ``[``-
        and ``]``-delimited attempt to access a local pith variable name
        uniquely identified by the passed 1-based index.

        Parameters
        ----------
        pith_index : int
            1-based index suffixing the local pith variable name to be created,
            cached, and returned.

        Returns
        -------
        str
            Prospective name of this local pith variable.

        Raises
        ------
        AssertionError
            If either:

            * ``pith_index`` is *not* an integer.
            * ``pith_index`` is a **negative integer** (i.e., less than 0).
        '''
        assert isinstance(pith_index, int), f'{repr(pith_index)} not integer.'
        assert pith_index >= 0, f'{pith_index} < 0.'
        # print(f'Generating indentation level {indent_level}...')

        # Prospective name of this local pith variable.
        pith_var_name = f'{VAR_NAME_PITH_PREFIX}{pith_index}'

        # Cache this name.
        self[pith_index] = pith_var_name

        # Return this name.
        return pith_var_name

# ....................{ MAPPINGS                           }....................
HINT_INDEX_TO_HINT_PLACEHOLDER = _HintIndexToHintPlaceholder()
'''
**Local type-checking placeholder cache singleton** (i.e., global dictionary
efficiently mapping from the 1-based index uniquely identifying each pith to the
corresponding type-checking placeholder substring).

Examples
--------
.. code-block:: pycon

   >>> from beartype._check.code.snip.codesnipcls import (
   ...     HINT_INDEX_TO_HINT_PLACEHOLDER)
   >>> HINT_INDEX_TO_HINT_PLACEHOLDER[1]
   '@[1)!'
   >>> HINT_INDEX_TO_HINT_PLACEHOLDER[2]
   '@[2)!'
'''


PITH_INDEX_TO_VAR_NAME = _PithIndexToVarName()
'''
**Local pith variable name cache singleton** (i.e., global dictionary
efficiently mapping from the 1-based index uniquely identifying each pith to the
name of a prospective local variable assigned that value in that body).

Examples
--------
.. code-block:: pycon

   >>> from beartype._check.code.snip.codesnipcls import PITH_INDEX_TO_VAR_NAME
   >>> PITH_INDEX_TO_VAR_NAME[1]
   '__beartype_pith_1'
   >>> PITH_INDEX_TO_VAR_NAME[2]
   '__beartype_pith_2'
'''

# ....................{ PRIVATE ~ constants                }....................
_HINT_PLACEHOLDER_PREFIX = '@['
'''
Prefix of each **placeholder hint child type-checking substring** (i.e.,
placeholder to be globally replaced by a Python code snippet type-checking the
current pith expression against the currently iterated child hint of the
currently visited parent hint).
'''


_HINT_PLACEHOLDER_SUFFIX = ')!'
'''
Suffix of each **placeholder hint child type-checking substring** (i.e.,
placeholder to be globally replaced by a Python code snippet type-checking the
current pith expression against the currently iterated child hint of the
currently visited parent hint).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype sanified type hint metadata dataclass** (i.e., class aggregating
*all* metadata returned by :mod:`beartype._check.convert.convmain` functions).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: [SPACE] Memoize the HintSane.__new__() or __init__() constructors. In
#theory, we dimly recall already defining a caching metaclass somewhere in the
#codebase. Perhaps we can simply leverage that to get this trivially done?
#
#Note, however, that keyword arguments will be an issue. We currently
#instantiate "HintSane" objects throughout the codebase by passing keyword
#arguments -- which clearly conflict with memoization. That said, preserving
#keyword argument passing would be *EXTREMELY* beneficial here. Without keyword
#arguments, we lose the flexibility that keyword arguments enable -- especially
#with respect to adding new keyword arguments at some future date.
#
#Perhaps that aforementioned caching metaclass could be augmented to support
#keyword arguments? That would still be better than nothing.
#FIXME: When memoizing, only memoize *CONDITIONALLY.* Notably, there exist two
#common cases here:
#* Context-free "HintSane" instances are initialized with *ONLY* a "hint". They
#  lack contextual metadata and are thus context-free. Unsurprisingly,
#  context-free "HintSane" instances are readily memoizable.
#* Contextual "HintSane" instances are initialized with both a "hint" and one or
#  more supplemental parameters supplying contextual metadata (e.g.,
#  "hint_recursable_to_depth", "typearg_to_hint"). They are *NOT* context-free.
#  Ergo, contextual "HintSane" instances are *NOT* readily memoizable. Don't
#  even bother wasting space or time attempting to do so.
#FIXME: Indeed, the above suggests the following:
#* Trivially conditionally memoize the HintSane.__new__() or __init__()
#  constructors *ONLY* when passed no optional keyword-only parameters (i.e.,
#  *ONLY* when passed the single "hint" parameter positionally).
#
#That's it. Shouldn't be that arduous and should speed things along. *shrug*

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeDecorHintSanifyException
from beartype.typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Set,
    Tuple,
    Union,
)
from beartype._data.typing.datatypingport import (
    Hint,
    Pep484612646TypeArgUnpackedToHint,
)
from beartype._data.kind.datakindmap import FROZENDICT_EMPTY
from beartype._util.kind.maplike.utilmapfrozen import FrozenDict
from beartype._util.utilobjmake import permute_object

# ....................{ HINTS                              }....................
FrozenDictHintToInt = Dict[Hint, int]
'''
PEP-compliant type hint matching any dictionary itself mapping from
PEP-compliant type hints to integers.

Caveats
-------
This hint currently erroneously matches mutable rather than immutable
dictionaries. While the latter would be preferable, Python lacks a builtin
immutable dictionary type and thus support for typing such types. So it goes.
'''

# ....................{ CLASSES                            }....................
#FIXME: Unit test us up, please.
class HintSane(object):
    '''
    **Sanified type hint metadata** (i.e., immutable and thus hashable object
    encapsulating *all* metadata returned by
    :mod:`beartype._check.convert.convmain` sanifiers after sanitizing a
    possibly PEP-noncompliant hint into a fully PEP-compliant hint).

    For efficiency, sanifiers only conditionally return this metadata for the
    proper subset of hints associated with this metadata; since most hints are
    *not* associated with this metadata, sanifiers typically only return a
    sanified type hint (rather than both that hint *and* this metadata).

    Caveats
    -------
    **Callers should avoid modifying this metadata.** For efficiency, this class
    does *not* explicitly prohibit modification of this metadata. Nonetheless,
    this class is implemented under the assumption that callers *never* modify
    this metadata. This metadata is effectively frozen. Any attempts to modify
    this metadata *will* induce nondeterminism throughout :mod:`beartype`,
    especially in memoized callables accepting and/or returning this metadata.

    Attributes
    ----------
    hint : Hint
        Type hint sanified (i.e., sanitized) from a possibly insane type hint
        into a hopefully sane type hint by a
        :mod:`beartype._check.convert.convmain` function.
    hint_recursable_to_depth : FrozenDictHintToInt
        Recursion guard implemented as a frozen dictionary mapping from each
        **transitive recursable parent hint** (i.e., direct or indirect parent
        hint of this sanified type hint such that that parent hint explicitly
        supports recursion) to that parent hint's **recursion depth** (i.e.,
        total number of times that parent hint has been visited during the
        current search from the root type hint down to this sanified type hint).
        If a subsequently visited child hint subscripting this hint already
        resides in this recursion guard, that child hint has already been
        visited by prior iteration and is thus recursive. Since recursive hints
        are valid (rather than constituting an unexpected error), the caller is
        expected to detect this use case and silently short-circuit infinite
        recursion by avoiding revisiting previously visited recursive hints.
    typearg_to_hint : Pep484612646TypeArgUnpackedToHint
        **Type parameter lookup table** (i.e., immutable dictionary mapping from
        the **type parameter** (i.e., :pep:`484`-compliant type variable or
        :pep:`646`-compliant unpacked type variable tuple) originally
        parametrizing the origins of all transitive parent hints of this hint if
        any to the corresponding child hints subscripting those parent hints).
        This table enables :func:`beartype.beartype` to efficiently reduce a
        proper subset of type parameters to non-type parameters at decoration
        time, including:

        * :pep:`484`- or :pep:`585`-compliant **subscripted generics.** For
          example, this table enables runtime type-checkers to reduce the
          semantically useless pseudo-superclass ``list[T]`` to the
          semantically useful pseudo-superclass ``list[int]`` at decoration time
          in the following example:

          .. code-block:: python

             class MuhGeneric[T](list[T]): pass

             @beartype
             def muh_func(muh_arg: MuhGeneric[int]) -> None: pass

        * :pep:`695`-compliant **subscripted type aliases.** For example, this
          table enables runtime type-checkers to reduce the semantically useless
          type hint ``muh_type_alias[float]`` to the semantically useful type
          hint ``float | int`` at decoration time in the following example:

          .. code-block:: python

             type muh_type_alias[T] = T | int

             @beartype
             def muh_func(muh_arg: muh_type_alias[float]) -> None: pass
    _hash : int
        Hash identifying this object, precomputed for efficiency.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently
    # called @beartype decorations. Slotting has been shown to reduce read and
    # write costs by approximately ~10%, which is non-trivial.
    __slots__ = (
        'hint',
        'hint_recursable_to_depth',
        'typearg_to_hint',
        '_hash',
    )


    # Squelch false negatives from mypy. This is absurd. This is mypy. See:
    #     https://github.com/python/mypy/issues/5941
    if TYPE_CHECKING:
        hint: Hint
        hint_recursable_to_depth: FrozenDictHintToInt
        typearg_to_hint: Pep484612646TypeArgUnpackedToHint


    _INIT_ARG_NAMES = frozenset((
        var_name
        for var_name in __slots__
        # Ignore private slotted instance variables defined above.
        if not var_name.startswith('_')
    ))
    '''
    Frozen set of the names of all parameters accepted by the :meth:`init`
    method, defined as the frozen set comprehension of all public slotted
    instance variables of this class.

    This frozen set enables efficient membership testing.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(
        self,

        # Mandatory parameters.
        hint: Hint,

        # Optional keyword-only parameters.
        *,
        hint_recursable_to_depth: FrozenDictHintToInt = FROZENDICT_EMPTY,
        typearg_to_hint: Pep484612646TypeArgUnpackedToHint = FROZENDICT_EMPTY,
    ) -> None:
        '''
        Initialize this sanified type hint metadata with the passed parameters.

        Parameters
        ----------
        hint : Hint
            Type hint sanified (i.e., sanitized) from a possibly insane type
            hint into a hopefully sane type hint by a
            :mod:`beartype._check.convert.convmain` function.
        hint_recursable_to_depth : FrozenDictHintToInt, default: FROZENDICT_EMPTY
            Recursion guard implemented as a frozen dictionary mapping from each
            **transitive recursable parent hint** (i.e., direct or indirect
            parent hint of this sanified type hint such that that parent hint
            explicitly supports recursion) to that parent hint's **recursion
            depth** (i.e., total number of times that parent hint has been
            visited during the current search from the root type hint down to
            this sanified type hint). Defaults to the empty frozen dictionary.
        typearg_to_hint : Pep484612646TypeArgUnpackedToHint, default: FROZENDICT_EMPTY
            **Type variable lookup table** (i.e., immutable dictionary mapping
            from the **type variables** (i.e., :pep:`484`-compliant
            :class:`typing.TypeVar` objects) originally parametrizing the
            origins of all transitive parent hints of this hint if any to the
            corresponding child hints subscripting those parent hints). Defaults
            to the empty frozen dictionary.

        See the class docstring for further details.
        '''
        assert isinstance(hint_recursable_to_depth, FrozenDict), (
            f'{repr(hint_recursable_to_depth)} not frozen dictionary.')
        assert isinstance(typearg_to_hint, FrozenDict), (
            f'{repr(typearg_to_hint)} not frozen dictionary.')

        # Classify all passed parameters as instance variables.
        self.hint = hint
        self.hint_recursable_to_depth = hint_recursable_to_depth
        self.typearg_to_hint = typearg_to_hint

        # Hash identifying this object, precomputed for efficiency.
        self._hash = hash((hint, hint_recursable_to_depth, typearg_to_hint))

    # ..................{ DUNDERS                            }..................
    def __hash__(self) -> int:
        '''
        Hash identifying this sanified type hint metadata.

        Returns
        -------
        int
            This hash.
        '''

        return self._hash


    def __eq__(self, other: object) -> bool:
        '''
        :data:`True` only if this sanified type hint metadata is equal to the
        passed arbitrary object.

        Parameters
        ----------
        other : object
            Arbitrary object to be compared for equality against this metadata.

        Returns
        -------
        Union[bool, type(NotImplemented)]
            Either:

            * If this other object is also sanified type hint metadata, either:

              * If these metadatum share equal instance variables, :data:`True`.
              * Else, :data:`False`.

            * Else, :data:`NotImplemented`.
        '''

        # Return either...
        return (
            # If this other object is also sanified hint metadata, true only
            # if these metadatum share the same instance variables;
            (
                self.hint == other.hint and
                self.hint_recursable_to_depth == other.hint_recursable_to_depth and
                self.typearg_to_hint == other.typearg_to_hint
            )
            if isinstance(other, HintSane) else
            # Else, this other object is *NOT* also sanified hint metadata. In
            # this case, the standard singleton informing Python that this
            # equality comparator fails to support this comparison.
            NotImplemented  # type: ignore[return-value]
        )


    def __repr__(self) -> str:
        '''
        Machine-readable representation of this metadata.
        '''

        # If this metadata is the ignorable "HINT_SANE_IGNORABLE" singleton,
        # trivially return the unqualified basename of this singleton for
        # debuggability, disambiguity, and readability.
        if self is HINT_SANE_IGNORABLE:
            return 'HINT_SANE_IGNORABLE'
        # Else, this metadata is *NOT* the ignorable "HINT_SANE_IGNORABLE" singleton.

        # Represent this metadata with just the minimal subset of metadata
        # needed to reasonably describe this metadata.
        return (
            f'{self.__class__.__name__}('
            f'hint={repr(self.hint)}, '
            f'hint_recursable_to_depth={repr(self.hint_recursable_to_depth)}, '
            f'typearg_to_hint={repr(self.typearg_to_hint)}'
            f')'
        )

    # ..................{ PERMUTERS                          }..................
    def permute_sane(self, **kwargs) -> 'HintSane':
        '''
        Shallow copy of this metadata such that each passed keyword parameter
        overwrites the instance variable of the same name in this copy.

        Parameters
        ----------
        Keyword parameters of the same name and type as instance variables of
        this object (e.g., ``hint: Hint``, ``typearg_to_hint:
        Pep484612646TypeArgUnpackedToHint``).

        Returns
        -------
        HintSane
            Shallow copy of this metadata such that each keyword parameter
            overwrites the instance variable of the same name in this copy.

        Raises
        ------
        _BeartypeDecorHintSanifyException
            If the name of any passed keyword parameter is *not* that of an
            existing instance variable of this object.
        '''

        # Set us up the permutation! Make your time!
        return permute_object(
            obj=self,
            init_arg_name_to_value=kwargs,
            init_arg_names=self._INIT_ARG_NAMES,
            exception_cls=_BeartypeDecorHintSanifyException,
        )

# ....................{ GLOBALS                            }....................
HINT_IGNORABLE = Any
'''
**Ignorable sanified type hint** (i.e., singleton :class:`.Any` type hint
encapsulated by the metadata to which *all* deeply or shallowly ignorable type
hints are reduced by :mod:`beartype._check.convert.convmain` sanifiers).
'''


HINT_SANE_IGNORABLE = HintSane(hint=HINT_IGNORABLE)
'''
**Ignorable sanified type hint metadata** (i.e., singleton :class:`.HintSane`
instance to which *all* deeply or shallowly ignorable type hints are reduced by
:mod:`beartype._check.convert.convmain` sanifiers).

This singleton enables callers to trivially differentiate ignorable from
unignorable hints. After sanification, if a hint is sanified to:

* Literally this singleton, then that hint is ignorable.
* Any other object, then that hint is unignorable.
'''


HINT_SANE_RECURSIVE = HintSane(hint=HINT_IGNORABLE)
'''
**Recursive sanified type hint metadata** (i.e., singleton :class:`.HintSane`
instance to which **deeply recursive type hints** (i.e., recursive type hints
whose reducers recursively expand to at least two levels of of recursion) are
reduced by :mod:`beartype._check.convert.convmain` sanifiers).

This singleton enables callers to trivially differentiate deeply recursive from
ignorable hints. While deeply recursive hints are ignorable in *most* contexts,
deeply recursive hints are unignorable in other contexts (e.g., when child hints
of parent unions). Differentiating between these two cases thus requires a
distinct singleton from the comparable and significantly more common
:data:`.HINT_SANE_IGNORABLE` singleton.

After sanification, if a hint is sanified to:

* Literally this singleton, then that hint is deeply recursive.
* Any other object, then that hint is *not* deeply recursive.
'''

# ....................{ HINTS                              }....................
HintOrSane = Union[Hint, HintSane]
'''
PEP-compliant type hint matching either a type hint *or* **sanified type hint
metadata** (i.e., :class:`.HintSane` object).
'''

# ....................{ HINTS ~ container                  }....................
DictHintSaneToAny = Dict[HintSane, Any]
'''
PEP-compliant type hint matching a dictionary mapping from keys that are
**sanified type hint metadata** (i.e., :class:`.HintSane` objects) to arbitrary
objects.
'''


IterableHintSane = Iterable[HintSane]
'''
PEP-compliant type hint matching an iterable of zero or more **sanified type
hint metadata** (i.e., :class:`.HintSane` objects).
'''


ListHintOrSane = List[HintOrSane]
'''
PEP-compliant type hint matching a list of zero or more items, each of which is
either a type hint *or* **sanified type hint metadata** (i.e.,
:class:`.HintSane` object).
'''


ListHintSane = List[HintSane]
'''
PEP-compliant type hint matching a list of zero or more **sanified type hint
metadata** (i.e., :class:`.HintSane` objects).
'''


SetHintSane = Set[HintSane]
'''
PEP-compliant type hint matching a set of zero or more **sanified type hint
metadata** (i.e., :class:`.HintSane` objects).
'''


TupleHintSane = Tuple[HintSane, ...]
'''
PEP-compliant type hint matching a tuple of zero or more **sanified type hint
metadata** (i.e., :class:`.HintSane` objects).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **abstract syntax tree (AST) singletons** (i.e., objects pertaining
to ASTs commonly required throughout this codebase, reducing space and time
consumption by preallocating widely used AST-centric objects).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from ast import (
    ClassDef,
    FunctionDef,
    Load,
    Store,
)

# ....................{ NODES                              }....................
NODE_CONTEXT_LOAD = Load()
'''
**Node context load singleton** (i.e., object suitable for passing as the
``ctx`` keyword parameter accepted by the ``__init__()`` method of various
abstract syntax tree (AST) node classes).
'''


NODE_CONTEXT_STORE = Store()
'''
**Node context store singleton** (i.e., object suitable for passing as the
``ctx`` keyword parameter accepted by the ``__init__()`` method of various
abstract syntax tree (AST) node classes).
'''

# ....................{ TYPES                              }....................
TYPES_NODE_LEXICAL_SCOPE = frozenset((
    ClassDef,
    FunctionDef,
))
'''
Frozen set of all **lexically scoping abstract syntax tree (AST) node types**
(i.e., types of all AST nodes whose declaration defines a new lexical scope).
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
:mod:`pytest` **object factories** (i.e., low-level test-specific utility
functions creating and returning objects, typically for use in higher-level
:mod:`pytest` fixtures defined elsewhere in this test suite).
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from collections.abc import (
    Iterable,
)

# ....................{ CONTEXTS                           }....................
def make_container_from_funcs(func_names: 'Iterable[str]') -> list:
    '''
    List created by iterating over the passed iterable and, for the
    fully-qualified name of each callable in that iterable, dynamically
    importing that callable, calling that callable, and adding the items of the
    container returned by that callable to this list.

    Parameters
    ----------
    func_names : Iterable[str]
        Iterable of the fully-qualified names of all callables to be called.

    Returns
    -------
    T
        Instance of this container, iteratively composed from the containers
        returned by these callables.
    '''
    assert isinstance(func_names, Iterable), f'{repr(func_names)} not iterable.'

    # Defer utility-specific imports.
    from beartype._util.module.utilmodimport import import_module_attr

    # Temporary list with which to compose this container.
    main_list = []

    # For the passed name of each callable...
    for func_name in func_names:
        assert isinstance(func_name, str), f'{repr(func_name)} not string.'

        # This callable dynamically imported as this name.
        func = import_module_attr(func_name)

        # List produced by calling this callable.
        func_list = func()

        # Append all items in this list to this larger list to be returned.
        main_list.extend(func_list)

    # Return this list.
    return main_list
